"""Parse Authors lines to extract author and affiliation data."""
import re
from itertools import dropwhile
from typing import Dict, Iterator, List, Tuple

from .tex2utf import tex2utf

PREFIX_MATCH = 'van|der|de|la|von|del|della|da|mac|ter|dem|di|vaziri'

"""
Takes data from an Author: line in the current arXiv abstract
file and returns a structured set of data:

 author_list_ptr = [
  [ author1_keyname, author1_firstnames, author1_suffix, affil1, affil2 ] ,
  [ author2_keyname, author2_firstnames, author1_suffix, affil1 ] ,
  [ author3_keyname, author3_firstnames, author1_suffix ]
         ]

Abstracted from Dienst software for OAI1 and other uses. This
routine should just go away when a better metadata structure is
adopted that deals with names and affiliations properly.

Must remember that there is at least one person one the archive
who has only one name, this should clearly be considered the key name.

Code originally written by Christina Scovel, Simeon Warner Dec99/Jan00
 2000-10-16 - separated.
 2000-12-07 - added support for suffix
 2003-02-14 - get surname prefixes from arXiv::Filters::Index [Simeon]
 2007-10-01 - created test script, some tidying [Simeon]
 2018-05-25 - Translated from Perl to Python [Brian C.]
"""


def parse_author_affil(authors: str) -> List[List[str]]:
    """
    Parse author line and returns an list of author and affiliation data.

    The list for each author will have at least three elements for
    keyname, firstname(s) and suffix. The keyname will always have content
    but the other strings might be empty strings if there is no firstname
    or suffix. Any additional elements after the first three are affiliations,
    there may be zero or more.

    Handling of prefix "XX collaboration" etc. is duplicated here and in
    arXiv::HTML::AuthorLink -- it shouldn't be. Likely should just be here.

    This routine is just a wrapper around the two parts that first split
    the authors line into parts, and then back propagate the affiliations.
    The first part is to be used along for display where we do not want
    to back propagate affiliation information.

    :param authors: string of authors from abs file or similar
    :return:
    Returns a structured set of data:
    author_list_ptr = [
       [ author1_keyname, author1_firstnames, author1_suffix, affil1, affil2 ],
       [ author2_keyname, author2_firstnames, author1_suffix, affil1 ] ,
       [ author3_keyname, author3_firstnames, author1_suffix ]
    ]
    """
    return _parse_author_affil_back_propagate(
        **_parse_author_affil_split(authors))


def _parse_author_affil_split(author_line: str) -> Dict:
    """
    Split author line into author and affiliation data.

    Take author line, tidy spacing and punctuation, and then split up into
    individual author an affiliation data. Has special cases to avoid splitting
    an initial collaboration name and records in $back_propagate_affiliation_to
    the fact that affiliations should not be back propagated to collaboration
    names.

    Does not handle multiple collaboration names.
    """
    if not author_line:
        return {'author_list': [], 'back_prop': 0}

    names: List[str] = split_authors(author_line)
    if not names:
        return {'author_list': [], 'back_prop': 0}

    names = _remove_double_commas(names)
    # get rid of commas at back
    namesIter: Iterator[str] = reversed(
        list(dropwhile(lambda x: x == ',', reversed(names))))
    # get rid of commas at front
    names = list(dropwhile(lambda x: x == ',', namesIter))

    # Extract all names (all parts not starting with comma or paren)
    names = list(map(_tidy_name, filter(
        lambda x: re.match('^[^](,]', x), names)))
    names = list(filter(lambda n: not re.match(
        r'^\s*et\.?\s+al\.?\s*', n, flags=re.IGNORECASE), names))

    (names, author_list,
     back_propagate_affiliations_to) = _collaboration_at_start(names)

    (enumaffils) = _enum_collaboration_at_end(author_line)

    # Split name into keyname and firstnames/initials.
    # Deal with different patterns in turn: prefixes, suffixes, plain
    # and single name.
    patterns = [('double-prefix',
                 r'^(.*)\s+(' + PREFIX_MATCH + r')\s(' +
                 PREFIX_MATCH + r')\s(\S+)$'),
                ('name-prefix-name',
                 r'^(.*)\s+(' + PREFIX_MATCH + r')\s(\S+)$'),
                ('name-name-prefix',
                 r'^(.*)\s+(\S+)\s(I|II|III|IV|V|Sr|Jr|Sr\.|Jr\.)$'),
                ('name-name',
                 r'^(.*)\s+(\S+)$'), ]

    # Now go through names in turn and try to get affiliations
    # to go with them
    for name in names:
        pattern_matches = ((mtype, re.match(m, name, flags=re.IGNORECASE))
                           for (mtype, m) in patterns)

        (mtype, match) = next(((mtype, m)
                               for (mtype, m) in pattern_matches
                               if m is not None), ('default', None))
        if match is None:
            author_entry = [name, '', '']
        elif mtype == 'double-prefix':
            s = '{} {} {}'.format(match.group(
                2), match.group(3), match.group(4))
            author_entry = [s, match.group(1), '']
        elif mtype == 'name-prefix-name':
            s = '{} {}'.format(match.group(2), match.group(3))
            author_entry = [s, match.group(1), '']
        elif mtype == 'name-name-prefix':
            author_entry = [match.group(2), match.group(1), match.group(3)]
        elif mtype == 'name-name':
            author_entry = [match.group(2), match.group(1), '']
        else:
            author_entry = [name, '', '']

        # search back in author_line for affiliation
        author_entry = _add_affiliation(
            author_line, enumaffils, author_entry, name)
        author_list.append(author_entry)

    return {'author_list': author_list,
            'back_prop': back_propagate_affiliations_to}


def parse_author_affil_utf(authors: str) -> List:
    """
    Call parse_author_affil() and do TeX to UTF conversion.

    Output structure is the same but should be in UTF and not TeX.
    """
    if not authors:
        return []
    return list(map(lambda author: list(map(tex2utf, author)),
                    parse_author_affil(authors)))


def _remove_double_commas(items: List[str]) -> List[str]:

    parts: List[str] = []
    last = ''
    for pt in items:
        if pt == ',' and last == ',':
            continue
        else:
            parts.append(pt)
            last = pt
    return parts


def _tidy_name(name: str) -> str:
    name = re.sub(r'\s\s+', ' ', name)  # also gets rid of CR
    # add space after dot (except in TeX)
    name = re.sub(r'(?<!\\)\.(\S)', r'. \g<1>', name)
    return name


def _collaboration_at_start(names: List[str]) \
        -> Tuple[List[str], List[List[str]], int]:
    """Perform special handling of collaboration at start."""
    author_list = []

    back_propagate_affiliations_to = 0
    while len(names) > 0:
        m = re.search(r'([a-z0-9\s]+\s+(collaboration|group|team))',
                      names[0], flags=re.IGNORECASE)
        if not m:
            break

        # Add to author list
        author_list.append([m.group(1), '', ''])
        back_propagate_affiliations_to += 1
        # Remove from names
        names.pop(0)
        # Also swallow and following comma or colon
        if names and (names[0] == ',' or names[0] == ':'):
            names.pop(0)

    return names, author_list, back_propagate_affiliations_to


def _enum_collaboration_at_end(author_line: str)->Dict:
    """Get separate set of enumerated affiliations from end of author_line."""
    # Now see if we have a separate set of enumerated affiliations
    # This is indicated by finding '(\s*('
    line_m = re.search(r'\(\s*\((.*)$', author_line)
    if not line_m:
        return {}

    enumaffils = {}
    affils = re.sub(r'\s*\)\s*$', '', line_m.group(1))

    # Now expect to have '1) affil1 (2) affil2 (3) affil3'
    for affil in affils.split('('):
        # Now expect `1) affil1 ', discard if no match
        m = re.match(r'^(\d+)\)\s*(\S.*\S)\s*$', affil)
        if m:
            enumaffils[m.group(1)] = re.sub(r'[\.,\s]*$', '', m.group(2))

    return enumaffils


def _add_affiliation(author_line: str,
                     enumaffils: Dict,
                     author_entry: List[str],
                     name: str) -> List:
    """
    Add author affiliation to author_entry if one is found in author_line.

    This should deal with these cases
    Smith B(labX) Smith B(1) Smith B(1, 2) Smith B(1 & 2) Smith B(1 and 2)
    """
    en = re.escape(name)
    namerex = r'{}\s*\(([^\(\)]+)'.format(en.replace(' ', 's*'))
    m = re.search(namerex, author_line, flags=re.IGNORECASE)
    if not m:
        return author_entry

    # Now see if we have enumerated references (just commas, digits, &, and)
    affils = m.group(1).rstrip().lstrip()
    affils = re.sub(r'(&|and)/,', ',', affils, flags=re.IGNORECASE)

    if re.match(r'^[\d,\s]+$', affils):
        for affil in affils.split(','):
            if affil in enumaffils:
                author_entry.append(enumaffils[affil])
    else:
        author_entry.append(affils)

    return author_entry


def _parse_author_affil_back_propagate(author_list: List[List[str]],
                                       back_prop: int) -> List[List[str]]:
    """Back propagate author affiliation.

    Take the author list structure generated by parse_author_affil_split(..)
    and propagate affiliation information backwards to preceeding author
    entries where none was give. Stop before entry $back_prop to avoid
    adding affiliation information to collaboration names.

    given, eg:
      a.b.first, c.d.second (affil)
    implies
      a.b.first (affil), c.d.second (affil)
    and in more complex cases:
      a.b.first, c.d.second (1), e.f.third, g.h.forth (2,3)
    implies
      a.b.first (1), c.d.second (1), e.f.third (2,3), g.h.forth (2,3)
    """
    last_affil: List[str] = []
    for x in range(len(author_list) - 1, max(back_prop - 1, -1), -1):
        author_entry = author_list[x]
        if len(author_entry) > 3:  # author has affiliation,store
            last_affil = author_entry
        elif last_affil:
            # author doesn't have affil but later one did => copy
            author_entry.extend(last_affil[3:])

    return author_list


def split_authors(authors: str) -> List:
    """
    Split author string into authors entity lists.

    Take an author line as a string and return a reference to a list of the
    different name and affiliation blocks. While this does normalize spacing
    and 'and', it is a key feature that the set of strings returned can be
    concatenated to reproduce the original authors line. This code thus
    provides a very graceful degredation for badly formatted authors lines, as
    the text at least shows up.
    """
    # split authors field into blocks with boundaries of ( and )
    if not authors:
        return []
    aus = re.split(r'(\(|\))', authors)
    aus = list(filter(lambda x: x != '', aus))

    blocks = []
    if len(aus) == 1:
        blocks.append(authors)
    else:
        c = ''
        depth = 0
        for bit in aus:
            if bit == '':
                continue
            if bit == '(':  # track open parentheses
                depth += 1
                if depth == 1:
                    blocks.append(c)
                    c = '('
                else:
                    c = c + bit
            elif bit == ')':  # track close parentheses
                depth -= 1
                c = c + bit
                if depth == 0:
                    blocks.append(c)
                    c = ''
                else:  # haven't closed, so keep accumulating
                    continue
            else:
                c = c + bit
        if c:
            blocks.append(c)

    listx = []

    for block in blocks:
        block = re.sub(r'\s+', ' ', block)
        if re.match(r'^\(', block):  # it is a comment
            listx.append(block)
        else:  # it is a name
            block = re.sub(r',?\s+(and|\&)\s', ',', block)
            names = re.split(r'(,|:)\s*', block)
            for name in names:
                if not name:
                    continue
                name = name.rstrip().lstrip()
                if name:
                    listx.append(name)

    # Recombine suffixes that were separated with a comma
    parts: List[str] = []
    for p in listx:
        if re.match(r'^(Jr\.?|Sr\.?\[IV]{2,})$', p) \
                and len(parts) >= 2 \
                and parts[-1] == ',' \
                and not re.match(r'\)$', parts[-2]):
            separator = parts.pop()
            last = parts.pop()
            recomb = "{}{} {}".format(last, separator, p)
            parts.append(recomb)
        else:
            parts.append(p)

    return parts



"""
arXiv group, archive and category definitions.

arXiv categories are arranged in a hierarchy where there are archives
(astro-ph, cs, math, etc.) that contain subject classes (astro-ph has subject
classes CO, GA, etc.). We now use the term category to refer to any archive or
archive.subject_class that one can submit to (so hep-th and math.IT are both
categories). No subject class can be in more than one archive. However, our
scientific advisors identify some categories that should appear in more than
one archive because they bridge major subject areas. Examples include math.MP
== math-ph and stat.TH = math.ST. These are called category aliases and the
idea is that any article classified in one of the aliases categories also
appears in the other (canonical), but that most of the arXiv code for display,
search, etc. does not need to understand the break with hierarchy.

"""
from typing import Optional
from .definitions import CATEGORIES, CATEGORIES_ACTIVE, ARCHIVES, \
    ARCHIVES_ACTIVE, ARCHIVES_SUBSUMED, CATEGORY_ALIASES
from .category import Category, Archive, Group


def get_category_display(category: str, canonical: bool = True) -> str:
    """
    Get the display name of an arXiv category.

    Parameters
    ----------
    category : str
        Category identifier, e.g. ``nlin.AO``.
    canonical : bool
        If True (default) and the category is subsumed, the display name for
        the canonical category will be returned instead.

    Returns
    -------
    str
        Display name for the category, e.g. ``Adaptation and Self-Organizing
        Systems (nlin.AO)``.
    """
    if canonical:
        return Category(category).canonical.display
    return Category(category).display


def get_archive_display(archive: str, canonical: bool = True) -> str:
    """
    Get the display name of an arXiv archive.

    Parameters
    ----------
    archive : str
        Archive identifier, e.g. ``astro-ph``.
    canonical : bool
        If True (default) and the archive is subsumed, the display name for
        the canonical archive will be returned instead.

    Returns
    -------
    str
        Display name for the category, e.g. ``Astrophysics (astro-ph)``.
    """
    if canonical:
        return Archive(archive).canonical.display
    return Archive(archive).display


def get_group_display(group: str) -> str:
    """
    Get the display name of an arXiv group.

    Parameters
    ----------
    group : str
        Group identifier, e.g. ``grp_math``.

    Returns
    -------
    str
        Display name for the group, e.g. ``Mathematics (grp_math)``.
    """
    return Group(group).display



"""Utility functions for arxiv.search."""
# pylint: disable=C0330

import re
from arxiv import taxonomy

__all__ = ('parse_arxiv_id', )

_archive = '|'.join([re.escape(key) for key in taxonomy.ARCHIVES.keys()])
"""string for use in Regex for all arXiv archives"""

_category = '|'.join([re.escape(key) for key in taxonomy.CATEGORIES.keys()])

_prefix = r'(?P<arxiv_prefix>ar[xX]iv:)'
"""
Attempt to catch the arxiv prefix in front of arxiv ids.

E.g. so that it can be included in the <a> tag anchor (ARXIVNG-1284).
"""

ARXIV_REGEX = (
    "^(ar[xX]iv:)?((?:(?:(?:%s)(?:[.][A-Z]{2})?/[0-9]{2}(?:0[1-9]|1[0-2])"
    "\\d{3}(?:[vV]\\d+)?))|(?:(?:[0-9]{2}(?:0[1-9]|1[0-2])[.]"
    "\\d{4,5}(?:[vV]\\d+)?)))$" % _category
)

OLD_STYLE_WITH_ARCHIVE = re.compile(
    r'(?:%s)?(?P<arxiv_id>(%s)\/\d{2}[01]\d{4}(v\d*)?)' % (_prefix, _archive),
    re.I
)

OLD_STYLE_WITH_CATEGORY = re.compile(
    r'(?:%s)?(?P<arxiv_id>(%s)\/\d{2}[01]\d{4}(v\d*)?)' % (_prefix, _category),
    re.I
)

OLD_STYLE = re.compile(
    r'(?:%s)?(?P<arxiv_id>(%s)\/\d{2}[01]\d{4}(v\d*)?)'
    % (_prefix, f'{_archive}|{_category}'),
    re.I
)

STANDARD = re.compile(
    r'(?<![\d=\.])(?:%s)?(?P<arxiv_id>\d{4}\.\d{4,5}(v\d*)?)'
    % (_prefix),
    re.I
)


def parse_arxiv_id(value: str) -> str:
    """
    Parse arxiv id from string.

    Raises `ValidationError` if no arXiv ID.
    """
    m = re.search(STANDARD, value)
    if m:
        return m.group('arxiv_id')
    m = re.search(OLD_STYLE, value)
    if m:
        return m.group('arxiv_id')
    raise ValueError('Not a valid arXiv ID')



"""arXiv license definitions."""


LICENSE_ICON_BASE_URI = '/icons/licenses'
LICENSES = {
    # key is the license URI
    'http://arxiv.org/licenses/nonexclusive-distrib/1.0/': {
        'label': 'arXiv.org perpetual, non-exclusive license to '
                 'distribute this article',
        'note': '(Minimal rights required by arXiv.org. Select this '
                'unless you understand the implications of other '
                'licenses)',
        'order': 1,
        'is_current': True,
    },
    'http://creativecommons.org/licenses/by/4.0/': {
        'label': 'Creative Commons Attribution license',
        'order': 5,
        'is_current': True,
        'icon_uri': f'{LICENSE_ICON_BASE_URI}/by-4.0.png'},

    'http://creativecommons.org/licenses/by-sa/4.0/': {
        'order': 6,
        'label': 'Creative Commons Attribution-ShareAlike license',
        'is_current': True,
        'icon_uri': f'{LICENSE_ICON_BASE_URI}/by-sa-4.0.png'},

    'http://creativecommons.org/licenses/by-nc-sa/4.0/': {
        'order': 7,
        'label': 'Creative Commons Attribution-Noncommercial-ShareAlike '
                 'license',
        'is_current': True,
        'icon_uri': f'{LICENSE_ICON_BASE_URI}/by-nc-sa-4.0.png'},

    'http://creativecommons.org/publicdomain/zero/1.0/': {
        'label': 'Creative Commons Public Domain Declaration',
        'order': 8,
        'is_current': True,
        'icon_uri': f'{LICENSE_ICON_BASE_URI}/zero-1.0.png'
    },

    'http://arxiv.org/licenses/assumed-1991-2003/': {
        'label': 'Assumed arXiv.org perpetual, non-exclusive license to '
                 'distribute this article for submissions made before '
                 'January 2004',
        'order': 9,
        'is_current': False,
    },

    'http://creativecommons.org/licenses/by/3.0/': {
        'label': 'Creative Commons Attribution license',
        'order': 2,
        'is_current': False,
        'icon_uri': f'{LICENSE_ICON_BASE_URI}/by-3.0.png'
    },

    'http://creativecommons.org/licenses/by-nc-sa/3.0/': {
        'label': 'Creative Commons Attribution-Noncommercial-ShareAlike '
                 'license',
        'order': 3,
        'is_current': False,
        'icon_uri': f'{LICENSE_ICON_BASE_URI}/by-nc-sa-3.0.png'
    },
    'http://creativecommons.org/licenses/publicdomain/': {
        'label': 'Creative Commons Public Domain Declaration',
        'note': '(Suitable for US government employees, for example)',
        'order': 4,
        'is_current': False,
        'icon_uri': f'{LICENSE_ICON_BASE_URI}/publicdomain.png'
    }
}
NO_LICENSE_TEXT = 'I do not certify that any of the above licenses apply'

ASSUMED_LICENSE_URI = 'http://arxiv.org/licenses/assumed-1991-2003/'

TRANSLATED_LICENSES = {
    'http://creativecommons.org/licenses/by/3.0/':
    'http://creativecommons.org/licenses/by/4.0/',
    'http://creativecommons.org/licenses/by-nc-sa/3.0/':
    'http://creativecommons.org/licenses/by-nc-sa/4.0/',
    'http://creativecommons.org/licenses/publicdomain/':
    'http://creativecommons.org/publicdomain/zero/1.0/'
}
"""Historical license to updated/current license (old URI: new URI)."""

CURRENT_LICENSES = {
    k: v for k, v in LICENSES.items()
    if 'order' in v and 'is_current' in v and v['is_current']
}

CURRENT_LICENSE_URIS = \
    sorted(CURRENT_LICENSES, key=lambda x: CURRENT_LICENSES[x]['order'])
"""Current license URIs by display order."""



r"""
Patterns and functions to detect arXiv ids and Urls in text.

Functions to detect arXiv ids, URLs and DOI in text.
Functions to transform them to <a> tags.

These were originally jinja filters but became a little too big
for that so they were split out and made more general so they didn't
rely on the Flask context.

There are several classes of patterns we want to match but there is
some overlap in these patterns. To avoid looking for and parsing HTML in each
jinja filter, detecting these patterns is combined.

So far we are looking for:
DOIs            DOI: 10.1145/0001234.1234567
arXiv IDS:      1234.12345 1234.12345v1 hep-ph1307.1843
HTTP URLs:      http://something.org/myPaper/1234.12345
FTP URLs:       ftp://example.com/files/1234.12345

Just matching for arXiv ids with \d{4}\.\d{4,5} will match several of
these. To deal with this we are priortizing the matches and
interupting once one is found.

We should probably match DOIs first because they are the source of a
lot of false positives for arxiv matches.

Updated 18 January, 2019: refactored to combine arXiv + DOI regexes with URL
matching and link generation provided by `bleach
<https://bleach.readthedocs.io>`_. The bleach URL regex is extended a bit to
handle FTP addresses.
"""
from typing import Optional, List, Pattern, Match, Tuple, Callable, \
    NamedTuple, Dict, Union, Mapping
import re
from functools import partial
from urllib.parse import quote, urlparse

from flask import url_for
from jinja2 import Markup, escape
import bleach

from arxiv import identifier
from . import clickthrough


Attrs = Dict[Union[str, Tuple[None, str]], str]
Callback = Callable[[Attrs, bool], Attrs]


def _without_group_names(pattern: Pattern) -> str:
    ptn: str = re.sub(r'\(\?P<[^>\!]+>', '(?:', pattern.pattern)
    return ptn


DOI = re.compile(   # '10.1145/0001234.1234567'
    r'(?P<doi>10.\d{4,9}/[-._;()/:A-Z0-9]+)',
    re.I
)
"""
Pattern for matching DOIs.

We should probably match DOIs first because they are the source of a
lot of false positives for arxiv matches.

Only using the most general express from
https://www.crossref.org/blog/dois-and-matching-regular-expressions/
"""

ARXIV_PATTERNS = [
    identifier.OLD_STYLE_WITH_ARCHIVE,  # math/0501233 hep-ph/0611734
    identifier.OLD_STYLE_WITH_CATEGORY,  # math.GR/0601136v3 math.GR/0601136
    identifier.STANDARD  # 1609.05068 1207.1234v1 1207.1234 1807.12345v12
]
ARXIV_ID = re.compile(
    "|".join([rf"(?:{_without_group_names(pattern)})"
              for pattern in ARXIV_PATTERNS]),
    re.IGNORECASE | re.VERBOSE | re.UNICODE
)

OKCHARS = r'([a-z0-9,_.\-+~:]|%[a-f0-9]*)'
"""Chacters that are acceptable during PATH, QUERY and ANCHOR parts"""

PATH = rf'(?P<PATH>(/{OKCHARS}*)+)?'
"""Regex for path part of URLs for use in urlize"""

FTP = re.compile(rf'(?P<url>(?:ftp://)({OKCHARS}|(@))*{PATH})', re.I)
"""Regex to match FTP URLs in text."""

IP_ADDRESS = (r'(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}'
              r'(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)')
"""Regex to match an IP address."""

TLDS = '|'.join(bleach.linkifier.TLDS)
PROTOCOLS = '|'.join(bleach.linkifier.html5lib_shim.allowed_protocols)
URL = re.compile(
    rf"""(?:{FTP.pattern})|
    (?:\(*  # Match any opening parentheses.
    \b(?<![@/])(?:(?:{PROTOCOLS}):/{{0,3}}(?:(?:\w+:)?\w+@)?)?  # http://
    (?:
        (?:([\w-]+\.)+(?:{TLDS}))   # Conventional URL
        |(?:{IP_ADDRESS})           # IP address
    )(?:\:[0-9]+)?(?!\.\w)\b        # Port
    (?:[/?][^\s\{{\}}\|\\\^\[\]`<>"]*)?)
        # /path/zz (excluding "unsafe" chars from RFC 1738,
        # except for # and ~, which happen in practice)
    """,
    re.IGNORECASE | re.VERBOSE | re.UNICODE
)


def arxiv_id_to_url(arxiv_id: str) -> str:
    """Generate an URL for an arXiv ID."""
    url: str = url_for('abs_by_id', paper_id=arxiv_id)
    return url


def url_for_doi(doi: str) -> str:
    """Generate an URL for a DOI."""
    quoted_doi = quote(doi, safe='/')
    return f'https://dx.doi.org/{quoted_doi}'


def clickthrough_url_for_doi(doi: str) -> str:
    """Generate a clickthrough URL for a DOI."""
    return clickthrough.clickthrough_url(url_for_doi(doi))


def _extend_class_attr(attrs: Attrs, new_class: str) -> Attrs:
    if (None, 'class') not in attrs:
        attrs[(None, 'class')] = ''
    attrs[(None, 'class')] = (attrs[(None, 'class')] + f' {new_class}').strip()
    return attrs


def _add_rel_external(attrs: Attrs, new: bool = False) -> Attrs:
    o = urlparse(attrs[(None, 'href')])
    if not o.netloc.split(':')[0].endswith('arxiv.org'):   # External link?
        attrs[(None, 'rel')] = 'external'
        attrs['_text'] = f'this {o.scheme} URL'    # Replaces the link text.
        _extend_class_attr(attrs, 'link-external')
    elif (None, 'data-arxiv-id') not in attrs \
            and (None, 'data-doi') not in attrs:
        attrs['_text'] = f'this {o.scheme} URL'    # Replaces the link text.
        _extend_class_attr(attrs, 'link-internal')
    return attrs


def _add_scheme_info(attrs: Attrs, new: bool = False) -> Attrs:
    o = urlparse(attrs[(None, 'href')])
    if (None, 'class') not in attrs:
        attrs[(None, 'class')] = ''
    _extend_class_attr(attrs, f'link-{o.scheme}')
    return attrs


def _handle_arxiv_url(attrs: Attrs, new: bool = False) -> Attrs:
    """
    Screen for reference to an arXiv e-print, and generate URL.

    If the :ref:`.ARXIV_ID` pattern is used, it will generate a link with the
    identifier as a the target. We need to intercept these links, and generate
    a real URL.
    """
    href = attrs[(None, 'href')]
    if '://' in href:   # http://1902.00123
        target = href.split('://', 1)[1]
        prefix = ''
    elif ':' in href:   # arxiv:1902.00123
        target = href.split(':', 1)[1]
        prefix = 'arXiv:'
    else:
        target = href
        prefix = ''
    if ARXIV_ID.match(target):
        attrs[(None, 'href')] = arxiv_id_to_url(target)
        attrs['_text'] = f'{prefix}{target}'
        attrs[(None, 'data-arxiv-id')] = target     # Add arxiv="<arxiv id>"
    return attrs


def _handle_doi_url(attrs: Attrs, new: bool = False) -> Attrs:
    """
    Screen for reference to a DOI, and generate a URL.

    If the :ref:`.DOI` pattern is used, it will generate a link with the
    doi as a the target. We need to intercept these links, and generate
    a real URL.
    """
    href = attrs[(None, 'href')]
    if '://' in href:
        target = href.split('://', 1)[1]
        prefix = ''
    elif ':' in href:
        target = href.split(':', 1)[1]
        prefix = 'doi:'
    else:
        target = href
        prefix = ''
    if DOI.match(target):
        attrs[(None, 'href')] = clickthrough_url_for_doi(target)
        attrs['_text'] = f'{prefix}{target}'
        attrs[(None, 'data-doi')] = target     # Add arxiv="<arxiv id>"
    return attrs


PATTERNS = {
    'doi': DOI,
    'url': URL,
    'arxiv_id': ARXIV_ID
}
ORDER = ['arxiv_id', 'doi', 'url']
SUPPORTED_KINDS = ORDER
"""Identifier types that we can currently match and convert to to URLs."""

DEFAULT_CALLBACKS: List[Callback] = [
    _handle_doi_url,
    _handle_arxiv_url,
    _add_rel_external,
    _add_scheme_info
]


def _get_pattern(kinds: List[str]) -> Pattern:
    return re.compile(
        '|'.join([rf'(?:{PATTERNS[kind].pattern})' for kind
                 in sorted(kinds, key=lambda kind: ORDER.index(kind))]),
        re.IGNORECASE | re.VERBOSE | re.UNICODE
    )


def _get_linker(kinds: List[str],
                callbacks: List[Callback] = DEFAULT_CALLBACKS) \
        -> Callable[[str], str]:
    linker: Callable[[str], str] = bleach.linkifier.Linker(
        callbacks=callbacks,
        skip_tags=['a'],
        url_re=_get_pattern(kinds)
    ).linkify
    return linker


def urlize(text: str, kinds: List[str] = SUPPORTED_KINDS) -> str:
    """
    Convert URLs and certain identifiers to HTML links.

    Parameters
    ----------
    text : str
        Some text containing identifiers and/or URLs.
    kinds : list
        Names (from :ref:`SUPPORTED_KINDS`) of identifiers to match.

    Returns
    -------
    str
        The passed text, with identifiers and/or URLs converted to HTML links.

    """
    return _get_linker(kinds)(text)


def urlizer(kinds: List[str] = SUPPORTED_KINDS) -> Callable[[str], str]:
    """
    Generate a function to convert tokens to links.

    Parameters
    ----------
    kinds : list
        Names (from :ref:`SUPPORTED_KINDS`) of identifiers to match and
        convert.

    Returns
    -------
    callable
        A function with signature ``(str) -> str`` that converts identifiers
        to HTML links.

    """
    return _get_linker(kinds)



from unittest import TestCase
from flask import Flask
from arxiv.base.factory import create_web_app


class TestBaseAppFactory(TestCase):
    """Tests for :mod:`arxiv.base.factory`."""

    def test_create_web_app(self):
        """:func:`.create_web_app` generates a :class:`.Flask` instance."""

        app = create_web_app()
        self.assertIsInstance(app, Flask)



"""
Provides base HTTP exceptions and exception handling.

arXiv-base exceptions inherit from `Werkzeug HTTPExceptions
<http://werkzeug.pocoo.org/docs/0.14/exceptions/>`_. In most cases it will
make sense to simply use Werkzeug's many HTTP exceptions directly. Custom
exceptions can also be defined by further subclassing those exceptions.

These handlers are attached to any Flask app passed to
:class:`arxiv.base.Base`. Thus any services that extends arXiv base need only
define (if needed) and raise exceptions.

Controllers should raise HTTP exceptions. The exception message (description)
will be rendered within the base template for that exception class, including
markup.
"""

from typing import Callable, List, Tuple

from werkzeug.exceptions import NotFound, Forbidden, Unauthorized, \
    MethodNotAllowed, RequestEntityTooLarge, BadRequest, InternalServerError, \
    HTTPException
from flask import render_template, make_response, Response

from http import HTTPStatus as status

_handlers = []


class ConfigurationError(InternalServerError):
    """Raised when a configuration parameter is missing or invalid."""


def handler(exception: type) -> Callable:
    """Generate a decorator to register a handler for an exception."""
    def deco(func: Callable) -> Callable:
        """Register a function as an exception handler."""
        _handlers.append((exception, func))
        return func
    return deco


def get_handlers() -> List[Tuple[type, Callable]]:
    """
    Get a list of registered exception handlers.

    Returns
    -------
    list
        List of (:class:`.HTTPException`, callable) tuples.

    """
    return _handlers


@handler(NotFound)
def handle_not_found(error: NotFound) -> Response:
    """Render the base 404 error page."""
    rendered = render_template("base/404.html", error=error,
                               pagetitle="404 Not Found")
    response: Response = make_response(rendered)
    response.status_code = status.NOT_FOUND
    return response


@handler(Forbidden)
def handle_forbidden(error: Forbidden) -> Response:
    """Render the base 403 error page."""
    rendered = render_template("base/403.html", error=error,
                               pagetitle="403 Forbidden")
    response: Response = make_response(rendered)
    response.status_code = status.FORBIDDEN
    return response


@handler(Unauthorized)
def handle_unauthorized(error: Unauthorized) -> Response:
    """Render the base 401 error page."""
    rendered = render_template("base/401.html", error=error,
                               pagetitle="401 Unauthorized")
    response: Response = make_response(rendered)
    response.status_code = status.UNAUTHORIZED
    return response


@handler(MethodNotAllowed)
def handle_method_not_allowed(error: MethodNotAllowed) -> Response:
    """Render the base 405 error page."""
    rendered = render_template("base/405.html", error=error,
                               pagetitle="405 Method Not Allowed")
    response: Response = make_response(rendered)
    response.status_code = status.METHOD_NOT_ALLOWED
    return response


@handler(RequestEntityTooLarge)
def handle_request_entity_too_large(error: RequestEntityTooLarge) -> Response:
    """Render the base 413 error page."""
    rendered = render_template("base/413.html", error=error,
                               pagetitle="413 Request Entity Too Large")
    response: Response = make_response(rendered)
    response.status_code = status.REQUEST_ENTITY_TOO_LARGE
    return response


@handler(BadRequest)
def handle_bad_request(error: BadRequest) -> Response:
    """Render the base 400 error page."""
    rendered = render_template("base/400.html", error=error,
                               pagetitle="400 Bad Request")
    response: Response = make_response(rendered)
    response.status_code = status.BAD_REQUEST
    return response


@handler(InternalServerError)
def handle_internal_server_error(error: InternalServerError) -> Response:
    """Render the base 500 error page."""
    rendered = render_template("base/500.html", error=error,
                               pagetitle="500 Internal Server Error")
    response: Response = make_response(rendered)
    response.status_code = status.INTERNAL_SERVER_ERROR
    return response



"""Tests for :mod:`.integration`."""

from unittest import TestCase
from .meta import MetaIntegration


class ContextSingleton:
    """Mimics context switching in Flask."""

    def __init__(self):
        """Make a place to put contextualized instances."""
        self.current_context = -1
        self.contexts = {}

    def set_context(self, context):
        """Change to a different context."""
        self.current_context = context

    def get_context(self):
        """Get the current context."""
        return self.current_context


class TestMetaIntegration(TestCase):
    """Tests for :class:`.MetaIntegration`."""

    @classmethod
    def setUpClass(cls):
        """Define a simple web service integration class."""
        cls.ctx = ContextSingleton()

        class TestIntegration(metaclass=MetaIntegration):
            """Test service integration class."""

            def __init__(self, context):
                self.context = context

            @classmethod
            def current_session(ccls):
                """Get or create an instance for the current context."""
                current_context = cls.ctx.get_context()
                if current_context not in cls.ctx.contexts:
                    cls.ctx.contexts[current_context] = ccls(current_context)
                return cls.ctx.contexts[current_context]

            def get_something(self):
                """Do something based on the instance state."""
                return self.context + self.context

            def get_instance(self):
                """Get the instance that owns this bound method."""
                return self

        cls.TestIntegration = TestIntegration

    def setUp(self):
        """Clear all contexts."""
        self.ctx.contexts.clear()

    def test_get_instance(self):
        """Can work with instances of the class directly."""
        instance = self.TestIntegration('bar')
        self.assertIsInstance(instance, self.TestIntegration)
        self.assertEqual(instance.context, 'bar')
        self.assertEqual(instance.get_something(), 'barbar')

    def test_get_contextualized_instance(self):
        """Can get a contextualized instance."""
        self.ctx.set_context(2)
        instance = self.TestIntegration.current_session()
        self.assertIsInstance(instance, self.TestIntegration)
        self.assertEqual(instance.context, 2)
        self.assertEqual(instance.get_something(), 4)

        self.ctx.set_context(3)
        second_instance = self.TestIntegration.current_session()
        self.assertIsInstance(second_instance, self.TestIntegration)
        self.assertEqual(second_instance.context, 3)
        self.assertEqual(second_instance.get_something(), 6)

        self.assertNotEqual(id(second_instance), id(instance))

    def test_get_contextualized_method(self):
        """Get the method of the context-bound instance."""
        self.ctx.set_context(1)
        first_method = self.TestIntegration.get_instance

        self.ctx.set_context(2)
        second_method = self.TestIntegration.get_instance

        self.ctx.set_context(1)
        third_method = self.TestIntegration.get_instance

        self.assertNotEqual(id(first_method()), id(second_method()),
                            "These methods belong to different instances"
                            " because they were obtained in different"
                            " contexts.")
        self.assertEqual(id(first_method()), id(third_method()),
                         "These methods belong to the same instance because"
                         " they were obtained in the same context.")



"""Category and archive definitions."""

from datetime import date

GROUPS = {
    'grp_physics': {
        'name': 'Physics',
        'start_year': 1991,
        'default_archive': 'hep-th'
    },
    'grp_math': {
        'name': 'Mathematics',
        'start_year': 1992,
        'default_archive': 'math'
    },
    'grp_q-bio': {
        'name': 'Quantitative Biology',
        'start_year': 1992,
        'default_archive': 'q-bio'
    },
    'grp_cs': {
        'name': 'Computer Science',
        'start_year': 1993,
        'default_archive': 'cs'
    },
    'grp_test': {
        'name': 'Test',
        'start_year': 1995,
        'is_test': True
    },
    'grp_q-fin': {
        'name': 'Quantitative Finance',
        'start_year': 1997
    },
    'grp_stat': {
        'name': 'Statistics',
        'start_year': 1999
    },
    'grp_eess': {
        'name': 'Electrical Engineering and Systems Science',
        'start_year': 2017
    },
    'grp_econ': {
        'name': 'Economics',
        'start_year': 2017
    }
}
DEFAULT_GROUP = 'physics'

ARCHIVES = {
    'acc-phys': {
        'name': 'Accelerator Physics',
        'in_group': 'grp_physics',
        'start_date': date(1994, 11, 1),
        'end_date': date(1996, 9, 1)
    },
    'adap-org': {
        'name': 'Adaptation, Noise, and Self-Organizing Systems',
        'in_group': 'grp_physics',
        'start_date': date(1993, 3, 1),
        'end_date': date(1999, 12, 1)
    },
    'alg-geom': {
        'name': 'Algebraic Geometry',
        'in_group': 'grp_math',
        'start_date': date(1992, 2, 1),
        'end_date': date(1997, 12, 1)
    },
    'ao-sci': {
        'name': 'Atmospheric-Oceanic Sciences',
        'in_group': 'grp_physics',
        'start_date': date(1995, 2, 1),
        'end_date': date(1996, 9, 1)
    },
    'astro-ph': {
        'name': 'Astrophysics',
        'in_group': 'grp_physics',
        'start_date': date(1992, 4, 1)
    },
    'atom-ph': {
        'name': 'Atomic, Molecular and Optical Physics',
        'in_group': 'grp_physics',
        'start_date': date(1995, 9, 1),
        'end_date': date(1996, 9, 1)
    },
    'bayes-an': {
        'name': 'Bayesian Analysis',
        'in_group': 'grp_physics',
        'start_date': date(1995, 6, 1),
        'end_date': date(1996, 11, 1)
    },
    'chao-dyn': {
        'name': 'Chaotic Dynamics',
        'in_group': 'grp_physics',
        'start_date': date(1993, 1, 1),
        'end_date': date(1999, 12, 1)
    },
    'chem-ph': {
        'name': 'Chemical Physics',
        'in_group': 'grp_physics',
        'start_date': date(1994, 3, 1),
        'end_date': date(1996, 9, 1)
    },
    'cmp-lg': {
        'name': 'Computation and Language',
        'in_group': 'grp_cs',
        'start_date': date(1994, 4, 1),
        'end_date': date(1998, 9, 1)
    },
    'comp-gas': {
        'name': 'Cellular Automata and Lattice Gases',
        'in_group': 'grp_physics',
        'start_date': date(1993, 2, 1),
        'end_date': date(1999, 12, 1)
    },
    'cond-mat': {
        'name': 'Condensed Matter',
        'in_group': 'grp_physics',
        'start_date': date(1992, 4, 1)
    },
    'cs': {
        'name': 'Computer Science',
        'in_group': 'grp_cs',
        'start_date': date(1993, 1, 1)
    },
    'dg-ga': {
        'name': 'Differential Geometry',
        'in_group': 'grp_math',
        'start_date': date(1994, 6, 1),
        'end_date': date(1997, 12, 1)
    },
    'econ': {
        'name': 'Economics',
        'in_group': 'grp_econ',
        'start_date': date(2017, 9, 1)
    },
    'eess': {
        'name': 'Electrical Engineering and Systems Science',
        'in_group': 'grp_eess',
        'start_date': date(2017, 9, 1)
    },
    'funct-an': {
        'name': 'Functional Analysis',
        'in_group': 'grp_math',
        'start_date': date(1992, 4, 1),
        'end_date': date(1997, 12, 1)
    },
    'gr-qc': {
        'name': 'General Relativity and Quantum Cosmology',
        'in_group': 'grp_physics',
        'start_date': date(1992, 7, 1)
    },
    'hep-ex': {
        'name': 'High Energy Physics - Experiment',
        'in_group': 'grp_physics',
        'start_date': date(1994, 4, 1)
    },
    'hep-lat': {
        'name': 'High Energy Physics - Lattice',
        'in_group': 'grp_physics',
        'start_date': date(1992, 2, 1)
    },
    'hep-ph': {
        'name': 'High Energy Physics - Phenomenology',
        'in_group': 'grp_physics',
        'start_date': date(1992, 3, 1)
    },
    'hep-th': {
        'name': 'High Energy Physics - Theory',
        'in_group': 'grp_physics',
        'start_date': date(1991, 8, 1)
    },
    'math': {
        'name': 'Mathematics',
        'in_group': 'grp_math',
        'start_date': date(1992, 2, 1)
    },
    'math-ph': {
        'name': 'Mathematical Physics',
        'in_group': 'grp_physics',
        'start_date': date(1996, 9, 1)
    },
    'mtrl-th': {
        'name': 'Materials Theory',
        'in_group': 'grp_physics',
        'start_date': date(1994, 11, 1),
        'end_date': date(1996, 9, 1)
    },
    'nlin': {
        'name': 'Nonlinear Sciences',
        'in_group': 'grp_physics',
        'start_date': date(1993, 1, 1)
    },
    'nucl-ex': {
        'name': 'Nuclear Experiment',
        'in_group': 'grp_physics',
        'start_date': date(1994, 12, 1)
    },
    'nucl-th': {
        'name': 'Nuclear Theory',
        'in_group': 'grp_physics',
        'start_date': date(1992, 10, 1)
    },
    'patt-sol': {
        'name': 'Pattern Formation and Solitons',
        'in_group': 'grp_physics',
        'start_date': date(1993, 2, 1),
        'end_date': date(1999, 12, 1)
    },
    'physics': {
        'name': 'Physics',
        'in_group': 'grp_physics',
        'start_date': date(1996, 10, 1)
    },
    'plasm-ph': {
        'name': 'Plasma Physics',
        'in_group': 'grp_physics',
        'start_date': date(1995, 9, 1),
        'end_date': date(1996, 9, 1)
    },
    'q-alg': {
        'name': 'Quantum Algebra and Topology',
        'in_group': 'grp_math',
        'start_date': date(1994, 12, 1),
        'end_date': date(1997, 12, 1)
    },
    'q-bio': {
        'name': 'Quantitative Biology',
        'in_group': 'grp_q-bio',
        'start_date': date(2003, 9, 1)
    },
    'q-fin': {
        'name': 'Quantitative Finance',
        'in_group': 'grp_q-fin',
        'start_date': date(2008, 12, 1)
    },
    'quant-ph': {
        'name': 'Quantum Physics',
        'in_group': 'grp_physics',
        'start_date': date(1994, 12, 1)
    },
    'solv-int': {
        'name': 'Exactly Solvable and Integrable Systems',
        'in_group': 'grp_physics',
        'start_date': date(1993, 4, 1),
        'end_date': date(1999, 12, 1)
    },
    'stat': {
        'name': 'Statistics',
        'in_group': 'grp_stat',
        'start_date': date(2007, 4, 1)
    },
    'supr-con': {
        'name': 'Superconductivity',
        'in_group': 'grp_physics',
        'start_date': date(1994, 11, 1),
        'end_date': date(1996, 9, 1)
    },
    'test': {
        'name': 'Test',
        'in_group': 'grp_test',
        'is_test': True,
        'start_date': date(1995, 2, 1)
    }
}

ARCHIVES_ACTIVE = {key: value for key,
                   value in ARCHIVES.items()
                   if 'end_date' not in ARCHIVES[key]}

# defunct archives subsumed by categories
ARCHIVES_SUBSUMED = {
    'cmp-lg': 'cs.CL',
    'adap-org': 'nlin.AO',
    'comp-gas': 'nlin.CG',
    'chao-dyn': 'nlin.CD',
    'solv-int': 'nlin.SI',
    'patt-sol': 'nlin.PS',
    'alg-geom': 'math.AG',
    'dg-ga': 'math.DG',
    'funct-an': 'math.FA',
    'q-alg': 'math.QA',
    'mtrl-th': 'cond-mat.mtrl-sci',
    'supr-con': 'cond-mat.supr-con',
    'acc-phys': 'physics.acc-ph',
    'ao-sci': 'physics.ao-ph',
    'atom-ph': 'physics.atom-ph',
    'bayes-an': 'physics.data-an',
    'chem-ph': 'physics.chem-ph',
    'plasm-ph': 'physics.plasm-ph'
}

# Legacy bare-archive names used as primary and secondary
# categories for old submissions. yyyy-mm is last month allowed.
LEGACY_ARCHIVE_AS_PRIMARY = {
    'astro-ph': date(2008, 12, 1),
    'cond-mat': date(2004, 1, 1),
    'test': date(2010, 1, 1)
}
LEGACY_ARCHIVE_AS_SECONDARY = {
    'astro-ph': date(2008, 12, 1),
    'cond-mat': date(2004, 1, 1),
    'q-bio': date(2003, 8, 1),
    'test': date(2010, 1, 1)
}

CATEGORIES = {
    'acc-phys': {
        'name': 'Accelerator Physics',
        'in_archive': 'acc-phys',
        'is_active': False
    },
    'adap-org': {
        'name': 'Adaptation, Noise, and Self-Organizing Systems',
        'in_archive': 'adap-org',
        'is_active': False
    },
    'alg-geom': {
        'name': 'Algebraic Geometry',
        'in_archive': 'alg-geom',
        'is_active': False
    },
    'ao-sci': {
        'name': 'Atmospheric-Oceanic Sciences',
        'in_archive': 'ao-sci',
        'is_active': False
    },
    'astro-ph': {
        'name': 'Astrophysics',
        'in_archive': 'astro-ph',
        'is_active': False
    },
    'astro-ph.CO': {
        'name': 'Cosmology and Nongalactic Astrophysics',
        'in_archive': 'astro-ph',
        'description': 'Phenomenology of early universe, cosmic microwave '
                       'background, cosmological parameters, primordial '
                       'element abundances, extragalactic distance scale, '
                       'large-scale structure of the universe. Groups, '
                       'superclusters, voids, intergalactic medium. Particle '
                       'astrophysics: dark energy, dark matter, baryogenesis, '
                       'leptogenesis, inflationary models, reheating, '
                       'monopoles, WIMPs, cosmic strings, primordial black '
                       'holes, cosmological gravitational '
                       'radiation',
        'is_active': True
    },
    'astro-ph.EP': {
        'name': 'Earth and Planetary Astrophysics',
        'in_archive': 'astro-ph',
        'description': 'Interplanetary medium, planetary physics, planetary '
                       'astrobiology, extrasolar planets, comets, asteroids, '
                       'meteorites. Structure and formation of the solar '
                       'system',
        'is_active': True
    },
    'astro-ph.GA': {
        'name': 'Astrophysics of Galaxies',
        'in_archive': 'astro-ph',
        'description': 'Phenomena pertaining to galaxies or the Milky Way. '
                       'Star clusters, HII regions and planetary nebulae, the '
                       'interstellar medium, atomic and molecular clouds, '
                       'dust. Stellar populations. Galactic structure, '
                       'formation, dynamics. Galactic nuclei, bulges, disks, '
                       'halo. Active Galactic Nuclei, supermassive black '
                       'holes, quasars. Gravitational lens systems. The Milky '
                       'Way and its contents',
        'is_active': True
    },
    'astro-ph.HE': {
        'name': 'High Energy Astrophysical Phenomena',
        'in_archive': 'astro-ph',
        'description': 'Cosmic ray production, acceleration, propagation, '
                       'detection. Gamma ray astronomy and bursts, X-rays, '
                       'charged particles, supernovae and other explosive '
                       'phenomena, stellar remnants and accretion systems, '
                       'jets, microquasars, neutron stars, pulsars, black '
                       'holes',
        'is_active': True
    },
    'astro-ph.IM': {
        'name': 'Instrumentation and Methods for Astrophysics',
        'in_archive': 'astro-ph',
        'description': 'Detector and telescope design, experiment proposals. '
                       'Laboratory Astrophysics. Methods for data analysis, '
                       'statistical methods. Software, database '
                       'design',
        'is_active': True
    },
    'astro-ph.SR': {
        'name': 'Solar and Stellar Astrophysics',
        'in_archive': 'astro-ph',
        'description': 'White dwarfs, brown dwarfs, cataclysmic variables. '
                       'Star formation and protostellar systems, stellar '
                       'astrobiology, binary and multiple systems of stars, '
                       'stellar evolution and structure, coronas. Central '
                       'stars of planetary nebulae. Helioseismology, solar '
                       'neutrinos, production and detection of gravitational '
                       'radiation from stellar systems',
        'is_active': True
    },
    'atom-ph': {
        'name': 'Atomic, Molecular and Optical Physics',
        'in_archive': 'atom-ph',
        'is_active': False
    },
    'bayes-an': {
        'name': 'Bayesian Analysis',
        'in_archive': 'bayes-an',
        'is_active': False
    },
    'chao-dyn': {
        'name': 'Chaotic Dynamics',
        'in_archive': 'chao-dyn',
        'is_active': False
    },
    'chem-ph': {
        'name': 'Chemical Physics',
        'in_archive': 'chem-ph',
        'is_active': False
    },
    'cmp-lg': {
        'name': 'Computation and Language',
        'in_archive': 'cmp-lg',
        'is_active': False
    },
    'comp-gas': {
        'name': 'Cellular Automata and Lattice Gases',
        'in_archive': 'comp-gas',
        'is_active': False
    },
    'cond-mat': {
        'name': 'Condensed Matter',
        'in_archive': 'cond-mat',
        'is_active': False
    },
    'cond-mat.dis-nn': {
        'name': 'Disordered Systems and Neural Networks',
        'in_archive': 'cond-mat',
        'is_active': True
    },
    'cond-mat.mes-hall': {
        'name': 'Mesoscale and Nanoscale Physics',
        'in_archive': 'cond-mat',
        'description': 'Semiconducting nanostructures: quantum dots, wires, '
                       'and wells. Single electronics, spintronics, 2d '
                       'electron gases, quantum Hall effect, nanotubes, '
                       'graphene, plasmonic nanostructures',
        'is_active': True
    },
    'cond-mat.mtrl-sci': {
        'name': 'Materials Science',
        'in_archive': 'cond-mat',
        'description': 'Techniques, synthesis, characterization, structure.  '
                       'Structural phase transitions, mechanical properties, '
                       'phonons. Defects, adsorbates, interfaces',
        'is_active': True
    },
    'cond-mat.other': {
        'name': 'Other Condensed Matter',
        'in_archive': 'cond-mat',
        'description': 'Work in condensed matter that does not fit into the '
                       'other cond-mat classifications',
        'is_active': True
    },
    'cond-mat.quant-gas': {
        'name': 'Quantum Gases',
        'in_archive': 'cond-mat',
        'description': 'Ultracold atomic and molecular gases, Bose-Einstein '
                       'condensation, Feshbach resonances, spinor condensates,'
                       ' optical lattices, quantum simulation with cold atoms '
                       'and molecules, macroscopic interference phenomena',
        'is_active': True
    },
    'cond-mat.soft': {
        'name': 'Soft Condensed Matter',
        'in_archive': 'cond-mat',
        'description': 'Membranes, polymers, liquid crystals, glasses, '
                       'colloids, granular matter',
        'is_active': True
    },
    'cond-mat.stat-mech': {
        'name': 'Statistical Mechanics',
        'in_archive': 'cond-mat',
        'description': 'Phase transitions, thermodynamics, field theory, non-'
                       'equilibrium phenomena, renormalization group and '
                       'scaling, integrable models, turbulence',
        'is_active': True
    },
    'cond-mat.str-el': {
        'name': 'Strongly Correlated Electrons',
        'in_archive': 'cond-mat',
        'description': 'Quantum magnetism, non-Fermi liquids, spin liquids, '
                       'quantum criticality, charge density waves, metal-'
                       'insulator transitions',
        'is_active': True
    },
    'cond-mat.supr-con': {
        'name': 'Superconductivity',
        'in_archive': 'cond-mat',
        'description': 'Superconductivity: theory, models, experiment.  '
                       'Superflow in helium',
        'is_active': True
    },
    'cs.AI': {
        'name': 'Artificial Intelligence',
        'in_archive': 'cs',
        'description': 'Covers all areas of AI except Vision, Robotics, '
                       'Machine Learning, Multiagent Systems, and Computation '
                       'and Language (Natural Language Processing), which '
                       'have separate subject areas. In particular, includes '
                       'Expert Systems, Theorem Proving (although this may '
                       'overlap with Logic in Computer Science), Knowledge '
                       'Representation, Planning, and Uncertainty in AI. '
                       'Roughly includes material in ACM Subject Classes I.2.'
                       '0, I.2.1, I.2.3, I.2.4, I.2.8, and I.2.11.',
        'is_active': True
    },
    'cs.AR': {
        'name': 'Hardware Architecture',
        'in_archive': 'cs',
        'description': 'Covers systems organization and hardware architecture.'
                       ' Roughly includes material in ACM Subject Classes C.0,'
                       ' C.1, and C.5.',
        'is_active': True
    },
    'cs.CC': {
        'name': 'Computational Complexity',
        'in_archive': 'cs',
        'description': 'Covers models of computation, complexity classes, '
                       'structural complexity, complexity tradeoffs, upper '
                       'and lower bounds. Roughly includes material in ACM '
                       'Subject Classes F.1 (computation by abstract devices),'
                       ' F.2.3 (tradeoffs among complexity measures), and F.4.'
                       '3 (formal languages), although some material in '
                       'formal languages may be more appropriate for Logic in '
                       'Computer Science. Some material in F.2.1 and F.2.2, '
                       'may also be appropriate here, but is more likely to '
                       'have Data Structures and Algorithms as the primary '
                       'subject area.',
        'is_active': True
    },
    'cs.CE': {
        'name': 'Computational Engineering, Finance, and Science',
        'in_archive': 'cs',
        'description': 'Covers applications of computer science to the '
                       'mathematical modeling of complex systems in the '
                       'fields of science, engineering, and finance. Papers '
                       'here are interdisciplinary and applications-oriented, '
                       'focusing on techniques and tools that enable '
                       'challenging computational simulations to be performed,'
                       ' for which the use of supercomputers or distributed '
                       'computing platforms is often required. Includes '
                       'material in ACM Subject Classes J.2, J.3, and J.4 ('
                       'economics).',
        'is_active': True
    },
    'cs.CG': {
        'name': 'Computational Geometry',
        'in_archive': 'cs',
        'description': 'Roughly includes material in ACM Subject Classes I.3.'
                       '5 and F.2.2.',
        'is_active': True
    },
    'cs.CL': {
        'name': 'Computation and Language',
        'in_archive': 'cs',
        'description': 'Covers natural language processing. Roughly includes '
                       'material in ACM Subject Class I.2.7. Note that work '
                       'on artificial languages (programming languages, '
                       'logics, formal systems) that does not explicitly '
                       'address natural-language issues broadly construed ('
                       'natural-language processing, computational '
                       'linguistics, speech, text retrieval, etc.) is not '
                       'appropriate for this area.',
        'is_active': True
    },
    'cs.CR': {
        'name': 'Cryptography and Security',
        'in_archive': 'cs',
        'description': 'Covers all areas of cryptography and security '
                       'including authentication, public key cryptosytems, '
                       'proof-carrying code, etc. Roughly includes material '
                       'in ACM Subject Classes D.4.6 and E.3.',
        'is_active': True
    },
    'cs.CV': {
        'name': 'Computer Vision and Pattern Recognition',
        'in_archive': 'cs',
        'description': 'Covers image processing, computer vision, pattern '
                       'recognition, and scene understanding. Roughly '
                       'includes material in ACM Subject Classes I.2.10, I.4, '
                       'and I.5.',
        'is_active': True
    },
    'cs.CY': {
        'name': 'Computers and Society',
        'in_archive': 'cs',
        'description': 'Covers impact of computers on society, computer '
                       'ethics, information technology and public policy, '
                       'legal aspects of computing, computers and education. '
                       'Roughly includes material in ACM Subject Classes K.0, '
                       'K.2, K.3, K.4, K.5, and K.7.',
        'is_active': True
    },
    'cs.DB': {
        'name': 'Databases',
        'in_archive': 'cs',
        'description': 'Covers database management, datamining, and data '
                       'processing. Roughly includes material in ACM Subject '
                       'Classes E.2, E.5, H.0, H.2, and J.1.',
        'is_active': True
    },
    'cs.DC': {
        'name': 'Distributed, Parallel, and Cluster Computing',
        'in_archive': 'cs',
        'description': 'Covers fault-tolerance, distributed algorithms, '
                       'stabilility, parallel computation, and cluster '
                       'computing. Roughly includes material in ACM Subject '
                       'Classes C.1.2, C.1.4, C.2.4, D.1.3, D.4.5, D.4.7, '
                       'E.1.',
        'is_active': True
    },
    'cs.DL': {
        'name': 'Digital Libraries',
        'in_archive': 'cs',
        'description': 'Covers all aspects of the digital library design and '
                       'document and text creation. Note that there will be '
                       'some overlap with Information Retrieval (which is a '
                       'separate subject area). Roughly includes material in '
                       'ACM Subject Classes H.3.5, H.3.6, H.3.7, I.7.',
        'is_active': True
    },
    'cs.DM': {
        'name': 'Discrete Mathematics',
        'in_archive': 'cs',
        'description': 'Covers combinatorics, graph theory, applications of '
                       'probability. Roughly includes material in ACM Subject '
                       'Classes G.2 and G.3.',
        'is_active': True
    },
    'cs.DS': {
        'name': 'Data Structures and Algorithms',
        'in_archive': 'cs',
        'description': 'Covers data structures and analysis of algorithms. '
                       'Roughly includes material in ACM Subject Classes E.1, '
                       'E.2, F.2.1, and F.2.2.',
        'is_active': True
    },
    'cs.ET': {
        'name': 'Emerging Technologies',
        'in_archive': 'cs',
        'description': 'Covers approaches to information processing ('
                       'computing, communication, sensing) and bio-chemical '
                       'analysis based on alternatives to silicon CMOS-based '
                       'technologies, such as nanoscale electronic, photonic, '
                       'spin-based, superconducting, mechanical, bio-chemical '
                       'and quantum technologies (this list is not exclusive).'
                       ' Topics of interest include (1) building blocks for '
                       'emerging technologies, their scalability and adoption '
                       'in larger systems, including integration with '
                       'traditional technologies, (2) modeling, design and '
                       'optimization of novel devices and systems, (3) models '
                       'of computation, algorithm design and programming for '
                       'emerging technologies.',
        'is_active': True
    },
    'cs.FL': {
        'name': 'Formal Languages and Automata Theory',
        'in_archive': 'cs',
        'description': 'Covers automata theory, formal language theory, '
                       'grammars, and combinatorics on words. This roughly '
                       'corresponds to ACM Subject Classes F.1.1, and F.4.3. '
                       'Papers dealing with computational complexity should '
                       'go to cs.CC; papers dealing with logic should go to '
                       'cs.LO.',
        'is_active': True
    },
    'cs.GL': {
        'name': 'General Literature',
        'in_archive': 'cs',
        'description': 'Covers introductory material, survey material, '
                       'predictions of future trends, biographies, and '
                       'miscellaneous computer-science related material. '
                       'Roughly includes all of ACM Subject Class A, except '
                       'it does not include conference proceedings (which '
                       'will be listed in the appropriate subject area).',
        'is_active': True
    },
    'cs.GR': {
        'name': 'Graphics',
        'in_archive': 'cs',
        'description': 'Covers all aspects of computer graphics. Roughly '
                       'includes material in all of ACM Subject Class I.3, '
                       'except that I.3.5 is is likely to have Computational '
                       'Geometry as the primary subject area.',
        'is_active': True
    },
    'cs.GT': {
        'name': 'Computer Science and Game Theory',
        'in_archive': 'cs',
        'description': 'Covers all theoretical and applied aspects at the '
                       'intersection of computer science and game theory, '
                       'including work in mechanism design, learning in games '
                       '(which may overlap with Learning), foundations of '
                       'agent modeling in games (which may overlap with '
                       'Multiagent systems), coordination, specification and '
                       'formal methods for non-cooperative computational '
                       'environments. The area also deals with applications '
                       'of game theory to areas such as electronic commerce.',
        'is_active': True
    },
    'cs.HC': {
        'name': 'Human-Computer Interaction',
        'in_archive': 'cs',
        'description': 'Covers human factors, user interfaces, and '
                       'collaborative computing. Roughly includes material in '
                       'ACM Subject Classes H.1.2 and all of H.5, except for '
                       'H.5.1, which is more likely to have Multimedia as the '
                       'primary subject area.',
        'is_active': True
    },
    'cs.IR': {
        'name': 'Information Retrieval',
        'in_archive': 'cs',
        'description': 'Covers indexing, dictionaries, retrieval, content and '
                       'analysis. Roughly includes material in ACM Subject '
                       'Classes H.3.0, H.3.1, H.3.2, H.3.3, and H.3.4.',
        'is_active': True
    },
    'cs.IT': {
        'name': 'Information Theory',
        'in_archive': 'cs',
        'description': 'Covers theoretical and experimental aspects of '
                       'information theory and coding. Includes material in '
                       'ACM Subject Class E.4 and intersects with H.1.1.',
        'is_active': True
    },
    'cs.LG': {
        'name': 'Machine Learning',
        'in_archive': 'cs',
        'description': 'Covers machine learning and computational (PAC) '
                       'learning. Roughly includes material in ACM Subject '
                       'Class I.2.6.',
        'is_active': True
    },
    'cs.LO': {
        'name': 'Logic in Computer Science',
        'in_archive': 'cs',
        'description': 'Covers all aspects of logic in computer science, '
                       'including finite model theory, logics of programs, '
                       'modal logic, and program verification. Programming '
                       'language semantics should have Programming Languages '
                       'as the primary subject area. Roughly includes '
                       'material in ACM Subject Classes D.2.4, F.3.1, F.4.0, '
                       'F.4.1, and F.4.2; some material in F.4.3 (formal '
                       'languages) may also be appropriate here, although '
                       'Computational Complexity is typically the more '
                       'appropriate subject area.',
        'is_active': True
    },
    'cs.MA': {
        'name': 'Multiagent Systems',
        'in_archive': 'cs',
        'description': 'Covers multiagent systems, distributed artificial '
                       'intelligence, intelligent agents, coordinated '
                       'interactions. and practical applications. Roughly '
                       'covers ACM Subject Class I.2.11.',
        'is_active': True
    },
    'cs.MM': {
        'name': 'Multimedia',
        'in_archive': 'cs',
        'description': 'Roughly includes material in ACM Subject Class H.5.1.',
        'is_active': True
    },
    'cs.MS': {
        'name': 'Mathematical Software',
        'in_archive': 'cs',
        'description': 'Roughly includes material in ACM Subject Class G.4.',
        'is_active': True
    },
    'cs.NA': {
        'name': 'Numerical Analysis',
        'in_archive': 'cs',
        'description': 'Roughly includes material in ACM Subject Class G.1.',
        'is_active': True
    },
    'cs.NE': {
        'name': 'Neural and Evolutionary Computing',
        'in_archive': 'cs',
        'description': 'Covers neural networks, connectionism, genetic '
                       'algorithms, artificial life, adaptive behavior. '
                       'Roughly includes some material in ACM Subject Class C.'
                       '1.3, I.2.6, I.5.',
        'is_active': True
    },
    'cs.NI': {
        'name': 'Networking and Internet Architecture',
        'in_archive': 'cs',
        'description': 'Covers all aspects of computer communication networks,'
                       ' including network architecture and design, network '
                       'protocols, and internetwork standards (like TCP/IP). '
                       'Also includes topics, such as web caching, that are '
                       'directly relevant to Internet architecture and '
                       'performance. Roughly includes all of ACM Subject '
                       'Class C.2 except C.2.4, which is more likely to have '
                       'Distributed, Parallel, and Cluster Computing as the '
                       'primary subject area.',
        'is_active': True
    },
    'cs.OH': {
        'name': 'Other Computer Science',
        'in_archive': 'cs',
        'description': 'This is the classification to use for documents that '
                       'do not fit anywhere else.',
        'is_active': True
    },
    'cs.OS': {
        'name': 'Operating Systems',
        'in_archive': 'cs',
        'description': 'Roughly includes material in ACM Subject Classes D.4.'
                       '1, D.4.2., D.4.3, D.4.4, D.4.5, D.4.7, and D.4.9.',
        'is_active': True
    },
    'cs.PF': {
        'name': 'Performance',
        'in_archive': 'cs',
        'description': 'Covers performance measurement and evaluation, '
                       'queueing, and simulation. Roughly includes material '
                       'in ACM Subject Classes D.4.8 and K.6.2.',
        'is_active': True
    },
    'cs.PL': {
        'name': 'Programming Languages',
        'in_archive': 'cs',
        'description': 'Covers programming language semantics, language '
                       'features, programming approaches (such as object-'
                       'oriented programming, functional programming, logic '
                       'programming). Also includes material on compilers '
                       'oriented towards programming languages; other '
                       'material on compilers may be more appropriate in '
                       'Architecture (AR). Roughly includes material in ACM '
                       'Subject Classes D.1 and D.3.',
        'is_active': True
    },
    'cs.RO': {
        'name': 'Robotics',
        'in_archive': 'cs',
        'description': 'Roughly includes material in ACM Subject Class I.2.9.',
        'is_active': True
    },
    'cs.SC': {
        'name': 'Symbolic Computation',
        'in_archive': 'cs',
        'description': 'Roughly includes material in ACM Subject Class I.1.',
        'is_active': True
    },
    'cs.SD': {
        'name': 'Sound',
        'in_archive': 'cs',
        'description': 'Covers all aspects of computing with sound, and sound '
                       'as an information channel. Includes models of sound, '
                       'analysis and synthesis, audio user interfaces, '
                       'sonification of data, computer music, and sound '
                       'signal processing. Includes ACM Subject Class H.5.5, '
                       'and intersects with H.1.2, H.5.1, H.5.2, I.2.7, I.5.4,'
                       ' I.6.3, J.5, K.4.2.',
        'is_active': True
    },
    'cs.SE': {
        'name': 'Software Engineering',
        'in_archive': 'cs',
        'description': 'Covers design tools, software metrics, testing and '
                       'debugging, programming environments, etc. Roughly '
                       'includes material in all of ACM Subject Classes D.2, '
                       'except that D.2.4 (program verification) should '
                       'probably have Logics in Computer Science as the '
                       'primary subject area.',
        'is_active': True
    },
    'cs.SI': {
        'name': 'Social and Information Networks',
        'in_archive': 'cs',
        'description': 'Covers the design, analysis, and modeling of social '
                       'and information networks, including their '
                       'applications for on-line information access, '
                       'communication, and interaction, and their roles as '
                       'datasets in the exploration of questions in these and '
                       'other domains, including connections to the social '
                       'and biological sciences. Analysis and modeling of '
                       'such networks includes topics in ACM Subject classes '
                       'F.2, G.2, G.3, H.2, and I.2; applications in '
                       'computing include topics in H.3, H.4, and H.5; and '
                       'applications at the interface of computing and other '
                       'disciplines include topics in J.1--J.7. Papers on '
                       'computer communication systems and network protocols ('
                       'e.g. TCP/IP) are generally a closer fit to the '
                       'Networking and Internet Architecture (cs.NI) '
                       'category.',
        'is_active': True
    },
    'cs.SY': {
        'name': 'Systems and Control',
        'in_archive': 'cs',
        'description': 'This section includes theoretical and experimental '
                       'research covering all facets of automatic control '
                       'systems, having as focal point analysis and design '
                       'methods using tools of modeling, simulation and '
                       'optimization. Specific areas of research include '
                       'nonlinear, distributed, adaptive, stochastic and '
                       'robust control, hybrid and discrete event systems. '
                       'Application areas include automotive, aerospace, '
                       'process control, network control, biological systems, '
                       'multiagent and cooperative control, sensor networks, '
                       'control of cyberphysical and energy-related systems, '
                       'control of computing systems.',
        'is_active': True
    },
    'dg-ga': {
        'name': 'Differential Geometry',
        'in_archive': 'dg-ga',
        'is_active': False
    },
    'econ.EM': {
        'name': 'Econometrics',
        'in_archive': 'econ',
        'description': 'Econometric Theory, Micro-Econometrics, Macro-'
                       'Econometrics, Empirical Content of Economic Relations '
                       'discovered via New Methods, Methodological Aspects of '
                       'the Application of Statistical Inference to Economic '
                       'Data.',
        'is_active': True
    },
    'econ.GN': {
        'name': 'General Economics',
        'in_archive': 'econ',
        'description': 'General methodological and applied contributions to '
                       'economics.',
        'is_active': True
    },
    'econ.TH': {
        'name': 'Theoretical Economics',
        'in_archive': 'econ',
        'description': 'Includes theoretical contributions to Contract '
                       'Theory, Decision Theory, Game Theory, General '
                       'Equilibrium, Growth, Learning and Evolution, '
                       'Macroeconomics, Market and Mechanism Design, and '
                       'Social Choice.',
        'is_active': True
    },
    'eess.AS': {
        'name': 'Audio and Speech Processing',
        'in_archive': 'eess',
        'description': 'Theory and methods for processing signals '
                       'representing audio, speech, and language, and their '
                       'applications. This includes analysis, synthesis, '
                       'enhancement, transformation, classification and '
                       'interpretation of such signals as well as the design, '
                       'development, and evaluation of associated signal '
                       'processing systems. Machine learning and pattern '
                       'analysis applied to any of the above areas is also '
                       'welcome.  Specific topics of interest include: '
                       'auditory modeling and hearing aids; acoustic '
                       'beamforming and source localization; classification '
                       'of acoustic scenes; speaker separation; active noise '
                       'control and echo cancellation; enhancement; de-'
                       'reverberation; bioacoustics; music signals analysis, '
                       'synthesis and modification; music information '
                       'retrieval;  audio for multimedia and joint audio-'
                       'video processing; spoken and written language '
                       'modeling, segmentation, tagging, parsing, '
                       'understanding, and translation; text mining; speech '
                       'production, perception, and psychoacoustics; speech '
                       'analysis, synthesis, and perceptual modeling and '
                       'coding; robust speech recognition; speaker '
                       'recognition and characterization; deep learning, '
                       'online learning, and graphical models applied to '
                       'speech, audio, and language signals; and '
                       'implementation aspects ranging from system '
                       'architecture to fast algorithms.',
        'is_active': True
    },
    'eess.IV': {
        'name': 'Image and Video Processing',
        'in_archive': 'eess',
        'description': 'Theory, algorithms, and architectures for the '
                       'formation, capture, processing, communication, '
                       'analysis, and display of images, video, and '
                       'multidimensional signals in a wide variety of '
                       'applications. Topics of interest include: '
                       'mathematical, statistical, and perceptual image and '
                       'video modeling and representation; linear and '
                       'nonlinear filtering, de-blurring, enhancement, '
                       'restoration, and reconstruction from degraded, low-'
                       'resolution or tomographic data; lossless and lossy '
                       'compression and coding; segmentation, alignment, and '
                       'recognition; image rendering, visualization, and '
                       'printing; computational imaging, including ultrasound,'
                       ' tomographic and magnetic resonance imaging; and '
                       'image and video analysis, synthesis, storage, search '
                       'and retrieval.',
        'is_active': True
    },
    'eess.SP': {
        'name': 'Signal Processing',
        'in_archive': 'eess',
        'description': 'Theory, algorithms, performance analysis and '
                       'applications of signal and data analysis, including '
                       'physical modeling, processing, detection and '
                       'parameter estimation, learning, mining, retrieval, '
                       'and information extraction. The term "signal" '
                       'includes speech, audio, sonar, radar, geophysical, '
                       'physiological, (bio-) medical, image, video, and '
                       'multimodal natural and man-made signals, including '
                       'communication signals and data. Topics of interest '
                       'include: statistical signal processing, spectral '
                       'estimation and system identification; filter design, '
                       'adaptive filtering / stochastic learning; ('
                       'compressive) sampling, sensing, and transform-domain '
                       'methods including fast algorithms; signal processing '
                       'for machine learning and machine learning for signal '
                       'processing applications; in-network and graph signal '
                       'processing; convex and nonconvex optimization methods '
                       'for signal processing applications; radar, sonar, and '
                       'sensor array beamforming and direction finding; '
                       'communications signal processing; low power, multi-'
                       'core and system-on-chip signal processing; sensing, '
                       'communication, analysis and optimization for cyber-'
                       'physical systems such as power grids and the Internet '
                       'of Things.',
        'is_active': True
    },
    'funct-an': {
        'name': 'Functional Analysis',
        'in_archive': 'funct-an',
        'is_active': False
    },
    'gr-qc': {
        'name': 'General Relativity and Quantum Cosmology',
        'in_archive': 'gr-qc',
        'is_active': True
    },
    'hep-ex': {
        'name': 'High Energy Physics - Experiment',
        'in_archive': 'hep-ex',
        'is_active': True
    },
    'hep-lat': {
        'name': 'High Energy Physics - Lattice',
        'in_archive': 'hep-lat',
        'is_active': True
    },
    'hep-ph': {
        'name': 'High Energy Physics - Phenomenology',
        'in_archive': 'hep-ph',
        'is_active': True
    },
    'hep-th': {
        'name': 'High Energy Physics - Theory',
        'in_archive': 'hep-th',
        'is_active': True
    },
    'math-ph': {
        'name': 'Mathematical Physics',
        'in_archive': 'math-ph',
        'is_active': True
    },
    'math.AC': {
        'name': 'Commutative Algebra',
        'in_archive': 'math',
        'description': 'Commutative rings, modules, ideals, homological '
                       'algebra, computational aspects, invariant theory, '
                       'connections to algebraic geometry and '
                       'combinatorics',
        'is_active': True
    },
    'math.AG': {
        'name': 'Algebraic Geometry',
        'in_archive': 'math',
        'description': 'Algebraic varieties, stacks, sheaves, schemes, moduli '
                       'spaces, complex geometry, quantum '
                       'cohomology',
        'is_active': True
    },
    'math.AP': {
        'name': 'Analysis of PDEs',
        'in_archive': 'math',
        'description': 'Existence and uniqueness, boundary conditions, linear '
                       'and non-linear operators, stability, soliton theory, '
                       'integrable PDE\'s, conservation laws, qualitative '
                       'dynamics',
        'is_active': True
    },
    'math.AT': {
        'name': 'Algebraic Topology',
        'in_archive': 'math',
        'description': 'Homotopy theory, homological algebra, algebraic '
                       'treatments of manifolds',
        'is_active': True
    },
    'math.CA': {
        'name': 'Classical Analysis and ODEs',
        'in_archive': 'math',
        'description': 'Special functions, orthogonal polynomials, harmonic '
                       'analysis, ODE\'s, differential relations, calculus of '
                       'variations, approximations, expansions, asymptotics',
        'is_active': True
    },
    'math.CO': {
        'name': 'Combinatorics',
        'in_archive': 'math',
        'description': 'Discrete mathematics, graph theory, enumeration, '
                       'combinatorial optimization, Ramsey theory, '
                       'combinatorial game theory',
        'is_active': True
    },
    'math.CT': {
        'name': 'Category Theory',
        'in_archive': 'math',
        'description': 'Enriched categories, topoi, abelian categories, '
                       'monoidal categories, homological algebra',
        'is_active': True
    },
    'math.CV': {
        'name': 'Complex Variables',
        'in_archive': 'math',
        'description': 'Holomorphic functions, automorphic group actions and '
                       'forms, pseudoconvexity, complex geometry, analytic '
                       'spaces, analytic sheaves',
        'is_active': True
    },
    'math.DG': {
        'name': 'Differential Geometry',
        'in_archive': 'math',
        'description': 'Complex, contact, Riemannian, pseudo-Riemannian and '
                       'Finsler geometry, relativity, gauge theory, global '
                       'analysis',
        'is_active': True
    },
    'math.DS': {
        'name': 'Dynamical Systems',
        'in_archive': 'math',
        'description': 'Dynamics of differential equations and flows, '
                       'mechanics, classical few-body problems, iterations, '
                       'complex dynamics, delayed differential '
                       'equations',
        'is_active': True
    },
    'math.FA': {
        'name': 'Functional Analysis',
        'in_archive': 'math',
        'description': 'Banach spaces, function spaces, real functions, '
                       'integral transforms, theory of distributions, measure '
                       'theory',
        'is_active': True
    },
    'math.GM': {
        'name': 'General Mathematics',
        'in_archive': 'math',
        'description': 'Mathematical material of general interest, topics not '
                       'covered elsewhere',
        'is_active': True
    },
    'math.GN': {
        'name': 'General Topology',
        'in_archive': 'math',
        'description': 'Continuum theory, point-set topology, spaces with '
                       'algebraic structure, foundations, dimension theory, '
                       'local and global properties',
        'is_active': True,
    },
    'math.GR': {
        'name': 'Group Theory',
        'in_archive': 'math',
        'description': 'Finite groups, topological groups, representation '
                       'theory, cohomology, classification and structure',
        'is_active': True
    },
    'math.GT': {
        'name': 'Geometric Topology',
        'in_archive': 'math',
        'description': 'Manifolds, orbifolds, polyhedra, cell complexes, '
                       'foliations, geometric structures',
        'is_active': True
    },
    'math.HO': {
        'name': 'History and Overview',
        'in_archive': 'math',
        'description': 'Biographies, philosophy of mathematics, mathematics '
                       'education, recreational mathematics, communication of '
                       'mathematics',
        'is_active': True
    },
    'math.IT': {
        'name': 'Information Theory',
        'in_archive': 'math',
        'description': 'math.IT is an alias for cs.IT. Covers theoretical and '
                       'experimental aspects of information theory and '
                       'coding.',
        'is_active': True
    },
    'math.KT': {
        'name': 'K-Theory and Homology',
        'in_archive': 'math',
        'description': 'Algebraic and topological K-theory, relations with '
                       'topology, commutative algebra, and operator '
                       'algebras',
        'is_active': True
    },
    'math.LO': {
        'name': 'Logic',
        'in_archive': 'math',
        'description': 'Logic, set theory, point-set topology, formal '
                       'mathematics',
        'is_active': True
    },
    'math.MG': {
        'name': 'Metric Geometry',
        'in_archive': 'math',
        'description': 'Euclidean, hyperbolic, discrete, convex, coarse '
                       'geometry, comparisons in Riemannian geometry, '
                       'symmetric spaces',
        'is_active': True
    },
    'math.MP': {
        'name': 'Mathematical Physics',
        'in_archive': 'math',
        'description': 'Mathematical methods in quantum field theory, quantum '
                       'mechanics, statistical mechanics, condensed matter, '
                       'nuclear and atomic physics',
        'is_active': True
    },
    'math.NA': {
        'name': 'Numerical Analysis',
        'in_archive': 'math',
        'description': 'Numerical algorithms for problems in analysis and '
                       'algebra, scientific computation',
        'is_active': True
    },
    'math.NT': {
        'name': 'Number Theory',
        'in_archive': 'math',
        'description': 'Prime numbers, diophantine equations, analytic number '
                       'theory, algebraic number theory, arithmetic geometry, '
                       'Galois theory',
        'is_active': True
    },
    'math.OA': {
        'name': 'Operator Algebras',
        'in_archive': 'math',
        'description': 'Algebras of operators on Hilbert space, C^*-algebras, '
                       'von Neumann algebras, non-commutative geometry',
        'is_active': True
    },
    'math.OC': {
        'name': 'Optimization and Control',
        'in_archive': 'math',
        'description': 'Operations research, linear programming, control '
                       'theory, systems theory, optimal control, game '
                       'theory',
        'is_active': True
    },
    'math.PR': {
        'name': 'Probability',
        'in_archive': 'math',
        'description': 'Theory and applications of probability and stochastic '
                       'processes: e.g. central limit theorems, large '
                       'deviations, stochastic differential equations, models '
                       'from statistical mechanics, queuing theory',
        'is_active': True
    },
    'math.QA': {
        'name': 'Quantum Algebra',
        'in_archive': 'math',
        'description': 'Quantum groups, skein theories, operadic and '
                       'diagrammatic algebra, quantum field theory',
        'is_active': True
    },
    'math.RA': {
        'name': 'Rings and Algebras',
        'in_archive': 'math',
        'description': 'Non-commutative rings and algebras, non-associative '
                       'algebras, universal algebra and lattice theory, '
                       'linear algebra, semigroups',
        'is_active': True
    },
    'math.RT': {
        'name': 'Representation Theory',
        'in_archive': 'math',
        'description': 'Linear representations of algebras and groups, Lie '
                       'theory, associative algebras, multilinear algebra',
        'is_active': True
    },
    'math.SG': {
        'name': 'Symplectic Geometry',
        'in_archive': 'math',
        'description': 'Hamiltonian systems, symplectic flows, classical '
                       'integrable systems',
        'is_active': True
    },
    'math.SP': {
        'name': 'Spectral Theory',
        'in_archive': 'math',
        'description': 'Schrodinger operators, operators on manifolds, '
                       'general differential operators, numerical studies, '
                       'integral operators, discrete models, resonances, non-'
                       'self-adjoint operators, random operators/matrices',
        'is_active': True
    },
    'math.ST': {
        'name': 'Statistics Theory',
        'in_archive': 'math',
        'description': 'math.ST is an alias for stat.TH. Applied, '
                       'computational and theoretical statistics: e.g. '
                       'statistical inference, regression, time series, '
                       'multivariate analysis, data analysis, Markov chain '
                       'Monte Carlo, design of experiments, case studies',
        'is_active': True
    },
    'mtrl-th': {
        'name': 'Materials Theory',
        'in_archive': 'mtrl-th',
        'is_active': False
    },
    'nlin.AO': {
        'name': 'Adaptation and Self-Organizing Systems',
        'in_archive': 'nlin',
        'description': 'adaptation, self-organizing systems, statistical '
                       'physics, fluctuating systems, stochastic processes, '
                       'interacting particle systems, machine learning',
        'is_active': True
    },
    'nlin.CD': {
        'name': 'Chaotic Dynamics',
        'in_archive': 'nlin',
        'description': 'dynamical systems, chaos, quantum chaos, topological '
                       'dynamics, cycle expansions, turbulence, '
                       'propagation',
        'is_active': True
    },
    'nlin.CG': {
        'name': 'Cellular Automata and Lattice Gases',
        'in_archive': 'nlin',
        'description': 'computational methods, time series analysis, signal '
                       'processing, wavelets, lattice gases',
        'is_active': True
    },
    'nlin.PS': {
        'name': 'Pattern Formation and Solitons',
        'in_archive': 'nlin',
        'description': 'pattern formation, coherent structures, solitons',
        'is_active': True
    },
    'nlin.SI': {
        'name': 'Exactly Solvable and Integrable Systems',
        'in_archive': 'nlin',
        'description': 'exactly solvable systems, integrable PDEs, integrable '
                       'ODEs, Painleve analysis, integrable discrete maps, '
                       'solvable lattice models, integrable quantum '
                       'systems',
        'is_active': True
    },
    'nucl-ex': {
        'name': 'Nuclear Experiment',
        'in_archive': 'nucl-ex',
        'is_active': True
    },
    'nucl-th': {
        'name': 'Nuclear Theory',
        'in_archive': 'nucl-th',
        'is_active': True
    },
    'patt-sol': {
        'name': 'Pattern Formation and Solitons',
        'in_archive': 'patt-sol',
        'is_active': False
    },
    'physics.acc-ph': {
        'name': 'Accelerator Physics',
        'in_archive': 'physics',
        'is_active': True
    },
    'physics.ao-ph': {
        'name': 'Atmospheric and Oceanic Physics',
        'in_archive': 'physics',
        'is_active': True
    },
    'physics.app-ph': {
        'name': 'Applied Physics',
        'in_archive': 'physics',
        'is_active': True
    },
    'physics.atm-clus': {
        'name': 'Atomic and Molecular Clusters',
        'in_archive': 'physics',
        'is_active': True
    },
    'physics.atom-ph': {
        'name': 'Atomic Physics',
        'in_archive': 'physics',
        'is_active': True
    },
    'physics.bio-ph': {
        'name': 'Biological Physics',
        'in_archive': 'physics',
        'is_active': True
    },
    'physics.chem-ph': {
        'name': 'Chemical Physics',
        'in_archive': 'physics',
        'is_active': True
    },
    'physics.class-ph': {
        'name': 'Classical Physics',
        'in_archive': 'physics',
        'is_active': True
    },
    'physics.comp-ph': {
        'name': 'Computational Physics',
        'in_archive': 'physics',
        'is_active': True
    },
    'physics.data-an': {
        'name': 'Data Analysis, Statistics and Probability',
        'in_archive': 'physics',
        'is_active': True
    },
    'physics.ed-ph': {
        'name': 'Physics Education',
        'in_archive': 'physics',
        'is_active': True
    },
    'physics.flu-dyn': {
        'name': 'Fluid Dynamics',
        'in_archive': 'physics',
        'is_active': True
    },
    'physics.gen-ph': {
        'name': 'General Physics',
        'in_archive': 'physics',
        'is_active': True
    },
    'physics.geo-ph': {
        'name': 'Geophysics',
        'in_archive': 'physics',
        'is_active': True
    },
    'physics.hist-ph': {
        'name': 'History and Philosophy of Physics',
        'in_archive': 'physics',
        'is_active': True
    },
    'physics.ins-det': {
        'name': 'Instrumentation and Detectors',
        'in_archive': 'physics',
        'is_active': True
    },
    'physics.med-ph': {
        'name': 'Medical Physics',
        'in_archive': 'physics',
        'is_active': True
    },
    'physics.optics': {
        'name': 'Optics',
        'in_archive': 'physics',
        'is_active': True
    },
    'physics.plasm-ph': {
        'name': 'Plasma Physics',
        'in_archive': 'physics',
        'is_active': True
    },
    'physics.pop-ph': {
        'name': 'Popular Physics',
        'in_archive': 'physics',
        'is_active': True
    },
    'physics.soc-ph': {
        'name': 'Physics and Society',
        'in_archive': 'physics',
        'is_active': True
    },
    'physics.space-ph': {
        'name': 'Space Physics',
        'in_archive': 'physics',
        'is_active': True
    },
    'plasm-ph': {
        'name': 'Plasma Physics',
        'in_archive': 'plasm-ph',
        'is_active': False
    },
    'q-alg': {
        'name': 'Quantum Algebra and Topology',
        'in_archive': 'q-alg',
        'is_active': False
    },
    'q-bio': {
        'name': 'Quantitative Biology',
        'in_archive': 'q-bio',
        'is_active': False
    },
    'q-bio.BM': {
        'name': 'Biomolecules',
        'in_archive': 'q-bio',
        'description': 'DNA, RNA, proteins, lipids, etc.; molecular '
                       'structures and folding kinetics; molecular '
                       'interactions; single-molecule manipulation.',
        'is_active': True
    },
    'q-bio.CB': {
        'name': 'Cell Behavior',
        'in_archive': 'q-bio',
        'description': 'Cell-cell signaling and interaction; morphogenesis '
                       'and development; apoptosis; bacterial conjugation; '
                       'viral-host interaction; '
                       'immunology',
        'is_active': True
    },
    'q-bio.GN': {
        'name': 'Genomics',
        'in_archive': 'q-bio',
        'description': 'DNA sequencing and assembly; gene and motif finding; '
                       'RNA editing and alternative splicing; genomic '
                       'structure and processes (replication, transcription, '
                       'methylation, etc); mutational processes.',
        'is_active': True
    },
    'q-bio.MN': {
        'name': 'Molecular Networks',
        'in_archive': 'q-bio',
        'description': 'Gene regulation, signal transduction, proteomics, '
                       'metabolomics, gene and enzymatic networks',
        'is_active': True
    },
    'q-bio.NC': {
        'name': 'Neurons and Cognition',
        'in_archive': 'q-bio',
        'description': 'Synapse, cortex, neuronal dynamics, neural network, '
                       'sensorimotor control, behavior, '
                       'attention',
        'is_active': True
    },
    'q-bio.OT': {
        'name': 'Other Quantitative Biology',
        'in_archive': 'q-bio',
        'description': 'Work in quantitative biology that does not fit into '
                       'the other q-bio classifications',
        'is_active': True
    },
    'q-bio.PE': {
        'name': 'Populations and Evolution',
        'in_archive': 'q-bio',
        'description': 'Population dynamics, spatio-temporal and '
                       'epidemiological models, dynamic speciation, co-'
                       'evolution, biodiversity, foodwebs, aging; molecular '
                       'evolution and phylogeny; directed evolution; origin '
                       'of life',
        'is_active': True
    },
    'q-bio.QM': {
        'name': 'Quantitative Methods',
        'in_archive': 'q-bio',
        'description': 'All experimental, numerical, statistical and '
                       'mathematical contributions of value to biology',
        'is_active': True
    },
    'q-bio.SC': {
        'name': 'Subcellular Processes',
        'in_archive': 'q-bio',
        'description': 'Assembly and control of subcellular structures ('
                       'channels, organelles, cytoskeletons, capsules, etc.); '
                       'molecular motors, transport, subcellular localization;'
                       ' mitosis and meiosis',
        'is_active': True
    },
    'q-bio.TO': {
        'name': 'Tissues and Organs',
        'in_archive': 'q-bio',
        'description': 'Blood flow in vessels, biomechanics of bones, '
                       'electrical waves, endocrine system, tumor growth',
        'is_active': True
    },
    'q-fin.CP': {
        'name': 'Computational Finance',
        'in_archive': 'q-fin',
        'description': 'Computational methods, including Monte Carlo, PDE, '
                       'lattice and other numerical methods with applications '
                       'to financial modeling',
        'is_active': True
    },
    'q-fin.EC': {
        'name': 'Economics',
        'in_archive': 'q-fin',
        'description': 'q-fin.EC is an alias for econ.GN. Economics, '
                       'including micro and macro economics, international '
                       'economics, theory of the firm, labor economics, and '
                       'other economic topics outside finance',
        'is_active': True
    },
    'q-fin.GN': {
        'name': 'General Finance',
        'in_archive': 'q-fin',
        'description': 'Development of general quantitative methodologies '
                       'with applications in finance',
        'is_active': True
    },
    'q-fin.MF': {
        'name': 'Mathematical Finance',
        'in_archive': 'q-fin',
        'description': 'Mathematical and analytical methods of finance, '
                       'including stochastic, probabilistic and functional '
                       'analysis, algebraic, geometric and other methods',
        'is_active': True
    },
    'q-fin.PM': {
        'name': 'Portfolio Management',
        'in_archive': 'q-fin',
        'description': 'Security selection and optimization, capital '
                       'allocation, investment strategies and performance '
                       'measurement',
        'is_active': True
    },
    'q-fin.PR': {
        'name': 'Pricing of Securities',
        'in_archive': 'q-fin',
        'description': 'Valuation and hedging of financial securities, their '
                       'derivatives, and structured products',
        'is_active': True
    },
    'q-fin.RM': {
        'name': 'Risk Management',
        'in_archive': 'q-fin',
        'description': 'Measurement and management of financial risks in '
                       'trading, banking, insurance, corporate and other '
                       'applications',
        'is_active': True
    },
    'q-fin.ST': {
        'name': 'Statistical Finance',
        'in_archive': 'q-fin',
        'description': 'Statistical, econometric and econophysics analyses '
                       'with applications to financial markets and economic '
                       'data',
        'is_active': True
    },
    'q-fin.TR': {
        'name': 'Trading and Market Microstructure',
        'in_archive': 'q-fin',
        'description': 'Market microstructure, liquidity, exchange and '
                       'auction design, automated trading, agent-based '
                       'modeling and market-making',
        'is_active': True
    },
    'quant-ph': {
        'name': 'Quantum Physics',
        'in_archive': 'quant-ph',
        'is_active': True
    },
    'solv-int': {
        'name': 'Exactly Solvable and Integrable Systems',
        'in_archive': 'solv-int',
        'is_active': False
    },
    'stat.AP': {
        'name': 'Applications',
        'in_archive': 'stat',
        'description': 'Biology, Education, Epidemiology, Engineering, '
                       'Environmental Sciences, Medical, Physical Sciences, '
                       'Quality Control, Social Sciences',
        'is_active': True
    },
    'stat.CO': {
        'name': 'Computation',
        'in_archive': 'stat',
        'description': 'Algorithms, Simulation, Visualization',
        'is_active': True
    },
    'stat.ME': {
        'name': 'Methodology',
        'in_archive': 'stat',
        'description': 'Design, Surveys, Model Selection, Multiple Testing, '
                       'Multivariate Methods, Signal and Image Processing, '
                       'Time Series, Smoothing, Spatial Statistics, Survival '
                       'Analysis, Nonparametric and Semiparametric '
                       'Methods',
        'is_active': True
    },
    'stat.ML': {
        'name': 'Machine Learning',
        'in_archive': 'stat',
        'description': 'Classification, Graphical Models, High Dimensional '
                       'Inference',
        'is_active': True
    },
    'stat.OT': {
        'name': 'Other Statistics',
        'in_archive': 'stat',
        'description': 'Work in statistics that does not fit into the other '
                       'stat classifications',
        'is_active': True
    },
    'stat.TH': {
        'name': 'Statistics Theory',
        'in_archive': 'stat',
        'description': 'Asymptotics, Bayesian Inference, Decision Theory, '
                       'Estimation, Foundations, Inference, Testing',
        'is_active': True
    },
    'supr-con': {
        'name': 'Superconductivity',
        'in_archive': 'supr-con',
        'is_active': False
    },
    'test': {
        'name': 'Test',
        'in_archive': 'test',
        'is_active': False
    },
    'test.dis-nn': {
        'name': 'Test Disruptive Networks',
        'in_archive': 'test',
        'is_active': False
    },
    'test.mes-hall': {
        'name': 'Test Hall',
        'in_archive': 'test',
        'is_active': False
    },
    'test.mtrl-sci': {
        'name': 'Test Mtrl-Sci',
        'in_archive': 'test',
        'is_active': False
    },
    'test.soft': {
        'name': 'Test Soft',
        'in_archive': 'test',
        'is_active': False
    },
    'test.stat-mech': {
        'name': 'Test Mechanics',
        'in_archive': 'test',
        'is_active': False
    },
    'test.str-el': {
        'name': 'Test Electrons',
        'in_archive': 'test',
        'is_active': False
    },
    'test.supr-con': {
        'name': 'Test Superconductivity',
        'in_archive': 'test',
        'is_active': False
    },
}

CATEGORIES_ACTIVE = {key: value for key, value in CATEGORIES.items()
                     if 'is_active' in CATEGORIES[key] and
                     CATEGORIES[key]['is_active']}

CATEGORY_ALIASES = {
    'math.MP': 'math-ph',
    'stat.TH': 'math.ST',
    'math.IT': 'cs.IT',
    'q-fin.EC': 'econ.GN'
}
"""
Equivalences: category alias: canonical category

This model is based on the notion that only two categories may be
equivalent--not more. There would have to be some significant changes
to the (classic) code to support three-way equivalences.
"""



"""
Contants for HTTP status codes. Merely to increase code readability.

This was originally implemented using ``int`` values. We then switched to using
the IntEnums in :mod:`http`, which uses names without ``HTTP_[0-9]{3}_``
prefixes. This module will preserve the mapping to the built-in IntEnums for
projects that haven't switched yet.
"""

from http import HTTPStatus
import warnings

warnings.warn("These constants are deprecated. Use the built-in"
              " http.HTTPStatus instead.", DeprecationWarning)

HTTP_100_CONTINUE = HTTPStatus.CONTINUE
HTTP_101_SWITCHING_PROTOCOLS = HTTPStatus.SWITCHING_PROTOCOLS
HTTP_200_OK = HTTPStatus.OK
HTTP_201_CREATED = HTTPStatus.CREATED
HTTP_202_ACCEPTED = HTTPStatus.ACCEPTED
HTTP_203_NON_AUTHORITATIVE_INFORMATION = \
    HTTPStatus.NON_AUTHORITATIVE_INFORMATION
HTTP_204_NO_CONTENT = HTTPStatus.NO_CONTENT
HTTP_205_RESET_CONTENT = HTTPStatus.RESET_CONTENT
HTTP_206_PARTIAL_CONTENT = HTTPStatus.PARTIAL_CONTENT
HTTP_300_MULTIPLE_CHOICES = HTTPStatus.MULTIPLE_CHOICES
HTTP_301_MOVED_PERMANENTLY = HTTPStatus.MOVED_PERMANENTLY
HTTP_302_FOUND = HTTPStatus.FOUND
HTTP_303_SEE_OTHER = HTTPStatus.SEE_OTHER
HTTP_304_NOT_MODIFIED = HTTPStatus.NOT_MODIFIED
HTTP_305_USE_PROXY = HTTPStatus.USE_PROXY
HTTP_306_RESERVED = 306
HTTP_307_TEMPORARY_REDIRECT = HTTPStatus.TEMPORARY_REDIRECT
HTTP_400_BAD_REQUEST = HTTPStatus.BAD_REQUEST
HTTP_401_UNAUTHORIZED = HTTPStatus.UNAUTHORIZED
HTTP_402_PAYMENT_REQUIRED = HTTPStatus.PAYMENT_REQUIRED
HTTP_403_FORBIDDEN = HTTPStatus.FORBIDDEN
HTTP_404_NOT_FOUND = HTTPStatus.NOT_FOUND
HTTP_405_METHOD_NOT_ALLOWED = HTTPStatus.METHOD_NOT_ALLOWED
HTTP_406_NOT_ACCEPTABLE = HTTPStatus.NOT_ACCEPTABLE
HTTP_407_PROXY_AUTHENTICATION_REQUIRED = \
    HTTPStatus.PROXY_AUTHENTICATION_REQUIRED
HTTP_408_REQUEST_TIMEOUT = HTTPStatus.REQUEST_TIMEOUT
HTTP_409_CONFLICT = HTTPStatus.CONFLICT
HTTP_410_GONE = HTTPStatus.GONE
HTTP_411_LENGTH_REQUIRED = HTTPStatus.LENGTH_REQUIRED
HTTP_412_PRECONDITION_FAILED = HTTPStatus.PRECONDITION_FAILED
HTTP_413_REQUEST_ENTITY_TOO_LARGE = HTTPStatus.REQUEST_ENTITY_TOO_LARGE
HTTP_414_REQUEST_URI_TOO_LONG = HTTPStatus.REQUEST_URI_TOO_LONG
HTTP_415_UNSUPPORTED_MEDIA_TYPE = HTTPStatus.UNSUPPORTED_MEDIA_TYPE
HTTP_416_REQUESTED_RANGE_NOT_SATISFIABLE = \
    HTTPStatus.REQUESTED_RANGE_NOT_SATISFIABLE
HTTP_417_EXPECTATION_FAILED = HTTPStatus.EXPECTATION_FAILED
HTTP_428_PRECONDITION_REQUIRED = HTTPStatus.PRECONDITION_REQUIRED
HTTP_429_TOO_MANY_REQUESTS = HTTPStatus.TOO_MANY_REQUESTS
HTTP_431_REQUEST_HEADER_FIELDS_TOO_LARGE = \
    HTTPStatus.REQUEST_HEADER_FIELDS_TOO_LARGE
HTTP_500_INTERNAL_SERVER_ERROR = HTTPStatus.INTERNAL_SERVER_ERROR
HTTP_501_NOT_IMPLEMENTED = HTTPStatus.NOT_IMPLEMENTED
HTTP_502_BAD_GATEWAY = HTTPStatus.BAD_GATEWAY
HTTP_503_SERVICE_UNAVAILABLE = HTTPStatus.SERVICE_UNAVAILABLE
HTTP_504_GATEWAY_TIMEOUT = HTTPStatus.GATEWAY_TIMEOUT
HTTP_505_HTTP_VERSION_NOT_SUPPORTED = HTTPStatus.HTTP_VERSION_NOT_SUPPORTED
HTTP_511_NETWORK_AUTHENTICATION_REQUIRED = \
    HTTPStatus.NETWORK_AUTHENTICATION_REQUIRED



"""Dern simple support for sending multipart/alternative e-mails."""

from typing import Optional, Dict, List
import os
from email.message import EmailMessage
from email.utils import formatdate
import smtplib

from flask import Flask, Blueprint

from arxiv.base.globals import get_application_config

NOREPLY = 'noreply@arxiv.org'


def send(recipient: str, subject: str, text_body: str,
         html_body: Optional[str] = None, sender: Optional[str] = None,
         headers: Dict[str, str] = {}, cc_recipients: List[str] = [],
         bcc_recipients: List[str] = []) -> None:
    """
    Send an e-mail.

    If both ``text_body`` and ``html_body`` are provided, the e-mail will be
    sent as ``multipart/alternative``.

    Parameters
    ----------
    recipient : str
        The e-mail address of the recipient.
    subject : str
        The subject line.
    text_body : str
        Plain text content of the e-mail.
    html_body : str
        HTML content of the e-mail.
    sender : str or None
        The e-mail address of the sender. If ``None`` (default), the default
        sender will be loaded from the current config.
    headers : dict
        Extra headers to add to the e-mail.
    cc_recipients : list
        E-mail addresses that should be CC recipients.
    bcc_recipients : list
        E-mail addresses that should be BCC recipients.

    """
    _send(_write(recipient, subject, text_body, html_body=html_body,
                 sender=sender, headers=headers, cc_recipients=cc_recipients,
                 bcc_recipients=bcc_recipients),
          host=_get_smtp_hostname(),
          port=_get_smtp_port(),
          local_hostname=_get_local_hostname(),
          username=_get_smtp_username(),
          password=_get_smtp_password(),
          use_ssl=_use_ssl())


def _write(recipient: str, subject: str, text_body: str,
           html_body: Optional[str] = None, sender: Optional[str] = None,
           headers: Dict[str, str] = {}, cc_recipients: List[str] = [],
           bcc_recipients: List[str] = []) -> EmailMessage:
    if sender is None:
        sender = _get_default_sender()
    message = EmailMessage()
    message['Subject'] = subject
    message['From'] = sender
    message['To'] = recipient

    # For bleeping context, see https://bugs.python.org/issue28879.
    message['Date'] = formatdate()

    if cc_recipients:
        message['CC'] = ', '.join(cc_recipients)
    if bcc_recipients:
        message['BCC'] = ', '.join(bcc_recipients)
    message.set_content(text_body)
    if html_body:
        message.add_alternative(html_body, subtype='html')
    return message


def _send(message: EmailMessage, host: str = 'localhost', port: int = 0,
          local_hostname: Optional[str] = None, use_ssl: bool = False,
          username: Optional[str] = None,
          password: Optional[str] = None) -> None:
    SMTP = smtplib.SMTP_SSL if use_ssl else smtplib.SMTP
    with SMTP(host, port, local_hostname) as s:
        if username and password:
            s.login(username, password)
        s.send_message(message)


def _get_default_sender() -> str:
    return get_application_config().get('DEFAULT_SENDER', NOREPLY)


def _get_smtp_hostname() -> str:
    return get_application_config().get('SMTP_HOSTNAME', 'localhost')


def _get_smtp_username() -> str:
    return get_application_config().get('SMTP_USERNAME', 'foouser')


def _get_smtp_password() -> str:
    return get_application_config().get('SMTP_PASSWORD', 'foopass')


def _get_smtp_port() -> int:
    return int(get_application_config().get('SMTP_PORT', '0'))


def _get_local_hostname() -> Optional[str]:
    return get_application_config().get('SMTP_LOCAL_HOSTNAME', None)


def _use_ssl() -> bool:
    return bool(int(get_application_config().get('SMTP_SSL', '0')))


def init_app(app: Flask) -> None:
    template_folder = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                                   'templates')
    blueprint = Blueprint('mail', __name__, template_folder=template_folder)
    app.register_blueprint(blueprint)



"""Tests for arXiv taxonomy module."""
from unittest import TestCase
from .  import LICENSES, CURRENT_LICENSE_URIS, ASSUMED_LICENSE_URI


class TestLicense(TestCase):
    """Tests for the arXiv license definitions."""

    def test_current_license_uris(self):
        """Regression test for the current active licenses."""
        self.assertListEqual(
            CURRENT_LICENSE_URIS,
            [
                'http://arxiv.org/licenses/nonexclusive-distrib/1.0/',
                'http://creativecommons.org/licenses/by/4.0/',
                'http://creativecommons.org/licenses/by-sa/4.0/',
                'http://creativecommons.org/licenses/by-nc-sa/4.0/',
                'http://creativecommons.org/publicdomain/zero/1.0/'
            ],
            'current license URIs match'
        )

    def test_licenses_are_valid(self):
        """Test licenses contain required key/values."""
        for uri, license in LICENSES.items():
            self.assertIn('label', license)
            self.assertTrue(license['label'])
            self.assertIn('is_current', license)
            self.assertIsInstance(license['is_current'], bool)
            self.assertIn('order', license)
            self.assertIsInstance(license['order'], int)
            if 'icon_uri' in license:
                self.assertTrue(license['icon_uri'])
            if 'note' in license:
                self.assertTrue(license['note'])

    def test_assumed_license_is_valid(self):
        """Test assumed license is defined in licenses."""
        self.assertIn(ASSUMED_LICENSE_URI, LICENSES)



"""
Provides routes for verifying base templates and exception handling.

The blueprint instantiated here is **not** for use in a production application;
it is not attached by :class:`arxiv.base.Base`.
"""

from typing import Any, Tuple, Callable, Dict
from datetime import datetime
from flask import Blueprint, render_template, current_app, make_response, \
    Response, flash, url_for

from http import HTTPStatus as status
from arxiv.base.exceptions import NotFound, Forbidden, Unauthorized, \
    MethodNotAllowed, RequestEntityTooLarge, BadRequest, InternalServerError

from . import alerts

blueprint = Blueprint('ui', __name__, url_prefix='',
                      static_folder='./static_test')


@blueprint.route('/styleguide', methods=['GET'])
def test_page() -> Response:
    """Render the test page."""
    rendered = render_template("base/styleguide.html", pagetitle='Home')
    response: Response = make_response(rendered, status.OK)

    # Demonstrate flash alerts. To see these alerts, reload the page.
    help_url = url_for('help')
    alerts.flash_warning(f'This is a warning, see <a href="{help_url}">the'
                         f' docs</a> for more information',
                         title='Warning title', safe=True)
    alerts.flash_info('This is some info', title='Info title')
    alerts.flash_failure('This is a failure', title='Failure title')
    alerts.flash_success('This is a success', title='Success title')
    alerts.flash_warning('This is a warning that cannot be dismissed',
                           dismissable=False)
    return response


@blueprint.route('/macros', methods=['GET'])
def test_macros() -> Response:
    """Test the template macros."""
    context = {
        'arxiv_id': 'physics/9707012',
        'title': 'Supersymmetric partner chirping of Newtonian free damping',
        'authors': 'H.C. Rosu, J.L. Romero, J. Socorro',
        'abstract': "We connect the classical free damping cases by means of"
                    " Rosner's construction in supersymmetric quantum"
                    " mechanics. Starting with the critical damping, one"
                    " can obtain in the underdamping case a chirping of"
                    " instantaneous physical frequency $\omega ^{2}(t) \propto"
                    " \omega_{u}^{2}sech^2(\omega_{u}t)$, whereas in the"
                    " overdamped case the \"chirping\" is of the (unphysical)"
                    " type $\omega ^{2}(t)\propto\omega_{o}^{2}sec^{2}(\omega_"
                    "{o}t)$, where $\omega_{u}$ and $\omega_{o}$ are the"
                    " underdamped and overdamped frequency parameters,"
                    " respectively.\n This has an \"abstract\" linefeed.""",
        'comments': "This version (physics/9707012v2) was not stored by arXiv."
                    " A subsequent replacement was made before versioning was"
                    " introduced.",
        'primary_category': 'cond-mat.supr-con',
        'submitted_date': datetime.now(),
        'submission_history': [
            {'version': 1, 'submitted_date': datetime.now()},
            {'version': 2, 'submitted_date': datetime.now()},
            {'version': 3, 'submitted_date': datetime.now()},
        ],
        'version': 2,
        'doi': '10.1000/182',
        'pagetitle': 'Home'
    }
    response: Response = render_template("base/testmacros.html", **context)
    return response


@blueprint.route('/404', methods=['GET'])
def test_404() -> Response:
    """Test the 404 error page."""
    raise NotFound()


@blueprint.route('/403', methods=['GET'])
def test_403() -> Response:
    """Test the 403 error page."""
    raise Forbidden()


@blueprint.route('/401', methods=['GET'])
def test_401() -> Response:
    """Test the 401 error page."""
    raise Unauthorized()


@blueprint.route('/405', methods=['GET'])
def test_405() -> Response:
    """Test the 405 error page."""
    raise MethodNotAllowed()


@blueprint.route('/413', methods=['GET'])
def test_413() -> Response:
    """Test the 413 error page."""
    raise RequestEntityTooLarge()


@blueprint.route('/400', methods=['GET'])
def test_400() -> Response:
    """Test the 400 error page."""
    raise BadRequest()


@blueprint.route('/500', methods=['GET'])
def test_500() -> Response:
    """Test the 500 error page."""
    raise InternalServerError()



from unittest import TestCase
from flask import Flask

from http import HTTPStatus as status
from arxiv.base import Base
from arxiv.base.factory import create_web_app


class TestExceptionHandling(TestCase):
    """HTTPExceptions should be handled with custom templates."""

    def setUp(self):
        """Initialize an app and install :class:`.Base`."""
        self.app = create_web_app()
        self.client = self.app.test_client()

    def test_401(self):
        """A 401 response should be returned."""
        response = self.client.get('/401')
        self.assertEqual(response.status_code, status.UNAUTHORIZED)
        self.assertEqual(response.content_type, 'text/html; charset=utf-8')

    def test_404(self):
        """A 404 response should be returned."""
        response = self.client.get('/404')
        self.assertEqual(response.status_code, status.NOT_FOUND)
        self.assertEqual(response.content_type, 'text/html; charset=utf-8')

    def test_400(self):
        """A 400 response should be returned."""
        response = self.client.get('/400')
        self.assertEqual(response.status_code, status.BAD_REQUEST)
        self.assertEqual(response.content_type, 'text/html; charset=utf-8')

    def test_403(self):
        """A 403 response should be returned."""
        response = self.client.get('/403')
        self.assertEqual(response.status_code, status.FORBIDDEN)
        self.assertEqual(response.content_type, 'text/html; charset=utf-8')

    def test_413(self):
        """A 413 response should be returned."""
        response = self.client.get('/413')
        self.assertEqual(response.status_code,
                         status.REQUEST_ENTITY_TOO_LARGE)
        self.assertEqual(response.content_type, 'text/html; charset=utf-8')

    def test_500(self):
        """A 500 response should be returned."""
        response = self.client.get('/500')
        self.assertEqual(response.status_code,
                         status.INTERNAL_SERVER_ERROR)
        self.assertEqual(response.content_type, 'text/html; charset=utf-8')



"""
Provides a logger factory with reasonable defaults.

We need to be able to analyze application logs in a consistent way across all
arXiv services. This module provides a :func:`.getLogger`, which works just
like the builtin logging.getLogger, but with a simpler interface:
``getLogger(name: str, stream: IO)``.

It sets some defaults that should be applied consistently across all arXiv
services (e.g. date formatting, overall message structure), so that we can
parse application log messages in a consistent way.
"""
from typing import IO
import logging
import sys
from datetime import datetime
from pytz import timezone, utc

from flask import request

from arxiv.base.globals import get_application_config


class RequestFormatter(logging.Formatter):
    """Logging formatter that adds the current request ID to the log record."""

    def format(self, record: logging.LogRecord) -> str:
        """Attach the request ID from the request environ to the log record."""
        try:
            request_id = request.environ.get('REQUEST_ID', None)
        except RuntimeError as e:
            request_id = None
        record.requestid = request_id   # type: ignore
        if 'paperid' not in record.__dict__:
            record.paperid = 'null'     # type: ignore
        return super().format(record)


def getLogger(name: str, stream: IO = sys.stderr) -> logging.Logger:
    """
    Wrapper for :func:`logging.getLogger` that applies configuration.

    Parameters
    ----------
    name : str

    Returns
    -------
    :class:`logging.Logger`
    """
    logger = logging.getLogger(name)
    config = get_application_config()

    # Set the log level from the Flask app configuration.
    level = int(config.get('LOGLEVEL', logging.INFO))
    logger.setLevel(level)

    # Log messages should be in Eastern local time, for consistency with
    # classic CUL/Apache logs.
    tz = timezone(config.get('ARXIV_BUSINESS_TZ', 'US/Eastern'))
    logging.Formatter.converter = lambda *args: datetime.now(tz=tz).timetuple()

    # Set the formats for log messages and asctime. We instantiate our own
    # StreamHandler so that we can use RequestFormatter (above).
    fmt = ("application %(asctime)s - %(name)s - %(requestid)s"
           " - [arxiv:%(paperid)s] - %(levelname)s: \"%(message)s\"")
    datefmt = '%d/%b/%Y:%H:%M:%S %z'    # Used to format asctime.
    handler = logging.StreamHandler(stream)
    handler.setFormatter(RequestFormatter(fmt=fmt, datefmt=datefmt))
    logger.handlers = []    # Clear default handler(s).
    logger.addHandler(handler)
    return logger
